import logging

from django.test import TestCase

from munch.apps.campaigns.models import OptOut
from munch.apps.users.tests.factories import UserFactory
from munch.apps.campaigns.tests.factories import MailFactory
from munch.apps.campaigns.tests.factories import MessageFactory
from munch.apps.campaigns.tests.factories import PreviewMailFactory

from .forms import AbuseNotificationForm


logging.disable(logging.CRITICAL)


class TestAbuseForms(TestCase):
    def test_mail_resolution(self):
        user = UserFactory()
        message = MessageFactory(author=user)
        mail = MailFactory(message=message)
        f = AbuseNotificationForm({
            'mail': mail.identifier,
            'comments': 'test'})
        f.is_valid()
        f.save()

    def test_form_validation(self):
        user = UserFactory()
        message = MessageFactory(author=user)
        mail = MailFactory(message=message)
        f = AbuseNotificationForm({
            'mail': mail.identifier,
            'comments': ''})
        self.assertFalse(f.is_valid())


class TestAbuseURL(TestCase):
    def test_abuse_url(self):
        user = UserFactory()
        message = MessageFactory(author=user)
        mail = MailFactory(message=message)
        self.assertEqual(
            mail.abuse_url,
            'http://test.munch.example.com/abuse/report/{}/'.format(
                mail.identifier))


class TestAbuseReport(TestCase):
    def test_abuse_makes_optout(self):
        """ An abuse should create an optout for the given mail """
        user = UserFactory()
        message = MessageFactory(author=user)
        mail = MailFactory(message=message)
        response = self.client.post(
            'http://test.munch.example.com/abuse/report/{}/'.format(
                mail.identifier),
            {'mail': mail.identifier, 'comments': 'test'})
        self.assertEqual(response.status_code, 302)
        optout = OptOut.objects.get(identifier=mail.identifier)
        self.assertEqual(optout.origin, OptOut.BY_ABUSE)


class TestAbusePreview(TestCase):
    def setUp(self):
        self.user = UserFactory()
        self.message = MessageFactory(author=self.user)

    def test_abuse_page_not_preview(self):
        """ Report page displays no warning"""
        mail = MailFactory(message=self.message)
        response = self.client.get('/abuse/report/{}/'.format(mail.identifier))
        self.assertEqual(response.status_code, 200)
        self.assertNotIn(
            "from a preview message", response.content.decode())

    def test_abuse_page_preview(self):
        """ Report page displays a warning"""
        preview = PreviewMailFactory(message=self.message)
        response = self.client.get(
            '/abuse/report/{}/'.format(preview.identifier))
        self.assertEqual(response.status_code, 200)
        self.assertIn(
            "from a preview message", response.content.decode())



# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2016-09-23 13:16
from __future__ import unicode_literals

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('abuse', '0001_initial'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='abusenotification',
            name='comment',
        ),
        migrations.RemoveField(
            model_name='abusenotification',
            name='contact_mail',
        ),
        migrations.RemoveField(
            model_name='abusenotification',
            name='contact_phone',
        ),
        migrations.RemoveField(
            model_name='abusenotification',
            name='first_name',
        ),
        migrations.RemoveField(
            model_name='abusenotification',
            name='organization',
        ),
        migrations.RemoveField(
            model_name='abusenotification',
            name='second_name',
        ),
        migrations.AddField(
            model_name='abusenotification',
            name='comments',
            field=models.TextField(null=True, verbose_name='comments'),
        ),
        migrations.AddField(
            model_name='abusenotification',
            name='contact_email',
            field=models.CharField(blank=True, max_length=50, validators=[django.core.validators.EmailValidator()], verbose_name='contact email'),
        ),
        migrations.AddField(
            model_name='abusenotification',
            name='contact_name',
            field=models.CharField(blank=True, max_length=100, verbose_name='contact name'),
        ),
        migrations.AlterField(
            model_name='abusenotification',
            name='mail',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='campaigns.Mail', verbose_name='mail'),
        ),
    ]



from django.contrib import admin

from .models import OptOut


class OptOutAdmin(admin.ModelAdmin):
    list_display = ['address', 'creation_date', 'origin']
    search_fields = ['address', 'identifier']
    list_filter = ['origin']
    readonly_fields = ['creation_date']
    raw_id_fields = ['author', 'category']


admin.site.register(OptOut, OptOutAdmin)



from django.core.exceptions import ValidationError


def properties_schema_validator(value):
    from .models import PROP_TYPES

    if not isinstance(value, list):
        raise ValidationError("Must be a list")

    for contact_field in value:
        prop_name = contact_field.get('name', '').strip()
        if not prop_name:
            raise ValidationError("Every fields must have a name")
        prop_type = contact_field.get('type', None)
        if not prop_type:
            raise ValidationError("Every fields must have a type")
        if prop_type not in PROP_TYPES:
            raise ValidationError(
                '{} is not an authorized type'.format(prop_type))



from django.urls import reverse
from django.shortcuts import render
from django.http import HttpResponseRedirect
from django.http import HttpResponseNotFound
from django.http import HttpResponseNotAllowed
from django.core.exceptions import ObjectDoesNotExist

from munch.core.utils import get_mail_by_identifier
from munch.apps.optouts.models import OptOut
from munch.apps.campaigns.models import PreviewMail


def unsubscribe(request, identifier):
    is_test_mail = identifier.startswith('preview')
    if is_test_mail:
        try:
            mail = PreviewMail.objects.get(identifier=identifier)
        except PreviewMail.DoesNotExist:
            return HttpResponseNotFound()
    else:
        try:
            mail = get_mail_by_identifier(identifier)
        except ObjectDoesNotExist:
            return HttpResponseNotFound()

    context = {'mail': mail, 'is_test_mail': is_test_mail}

    if identifier.startswith('c-') and mail.message.external_optout:
        if request.method == 'GET':
            return render(
                request, 'optouts/unsubscribe_external.html', context)
        else:
            return HttpResponseNotAllowed(['GET'])
    else:
        if request.method == 'POST':
            optout_kwargs = {
                'category': mail.get_category(), 'author': mail.get_author()}
            if not is_test_mail:
                OptOut.objects.create_or_update(
                    identifier=mail.identifier,
                    address=mail.recipient,
                    origin=OptOut.BY_WEB,
                    **optout_kwargs)
            return HttpResponseRedirect(
                reverse('unsubscribed', kwargs={'identifier': identifier}))
        else:
            return render(request, 'optouts/unsubscribe.html', context)


def unsubscribed(request, identifier):
    if identifier.startswith('preview'):
        mail = PreviewMail.objects.get(identifier=identifier)
    else:
        mail = get_mail_by_identifier(identifier)
    return render(request, 'optouts/unsubscribed.html', {'mail': mail})



from django.conf import settings
from slimta.policy import QueuePolicy

from munch.core.mail.utils import mk_base64_uuid


class Add(QueuePolicy):
    """
    Add Munch User-Id and Message-Id headers
    """
    def apply(self, envelope):
        envelope.headers[settings.TRANSACTIONAL[
            'X_USER_ID_HEADER']] = str(envelope.user.pk)
        envelope.headers[settings.TRANSACTIONAL[
            'X_MESSAGE_ID_HEADER']] = mk_base64_uuid(prefix='t-')



from munch.core.utils.filters import DRFFilterSet

from ...models import MunchUser


"""
Handles all filtering/ordering for API viewsets.

Default ordering is :
- (if defined) first item of order_by
- (else) first item of fields
"""


class MunchUserFilter(DRFFilterSet):
    class Meta:
        model = MunchUser
        fields = ['identifier']
        order_by = True



import time
import json

import lxml.etree
from lxml.cssselect import CSSSelector
from django.core import signing
from django.utils.encoding import force_text
from django.utils.encoding import force_bytes
from django.utils.http import urlsafe_base64_encode
from django.utils.http import urlsafe_base64_decode


class WebKey:
    """ Webkey protects access to model resources by signature

    One creates a webkey from a resource, and gets a token. It can be  embeded
    in URL.

    Then, a new WebKey can be instanciated with that token that contains
    signature, timestamp  and the pk of the object.

    get_instance() will return the instance only if the signature is ok.

    """
    @classmethod
    def from_instance(cls, identifier):
        """
        :param identifier: any object identifier
        :rtype: WebKey
        """
        payload = json.dumps({'t': time.time(), 'identifier': identifier})
        token = signing.Signer().sign(urlsafe_base64_encode(
            force_bytes(payload)))
        return cls(token)

    def __init__(self, token):
        self.token = token

    def get_identifier(self):
        """ Gets the instance, if token is valid.

        :param klass: the class (must be model.Model instance)
        May throw DoesNotExist
        :returns: instance if signature is ok, None else.
        """
        try:
            payload = signing.Signer().unsign(self.token)
        except signing.BadSignature:
            return None
        else:
            d = json.loads(force_text(urlsafe_base64_decode(payload)))
            return d.get('identifier')


def get_msg_links(html):
    from munch.apps.tracking.models import LinkMap

    links_selector = CSSSelector('a')
    links = []

    try:
        doc = lxml.etree.HTML(html).getroottree()
    except lxml.etree.XMLSyntaxError:
        return {}

    for link in links_selector(doc):
        original_url = link.get('href') or ''
        original_url = original_url.strip()

        if original_url.startswith('http'):
            if original_url not in links:
                links.append(original_url)

    # Store the links_map with IDs as keys
    links_maps = LinkMap.objects.bulk_create(
        [LinkMap(link=link) for link in links])

    return {lm.identifier: lm.link for lm in links_maps}



from datetime import timedelta

from celery import task
from django.utils import timezone
from celery.utils.log import get_task_logger

from .models import SendingDomain
from .fields import DomainCheckField

log = get_task_logger(__name__)


@task
def run_domains_validation(statuses):
    """
        Task that iterate over all domains with specified statuses
        in order to re-check them.
    """
    for field in ['dkim', 'app_domain']:
        if statuses == [DomainCheckField.OK]:
            filters = {'{}_status'.format(field): DomainCheckField.OK}
        else:
            filters = {
                '{}_status__in'.format(field): statuses,
                '{}_status_date__gt'.format(
                    field): timezone.now() - timedelta(days=2)}

        qs = SendingDomain.objects.filter(**filters).only('id')
        log.info(
            'Running sending domain "{}" validations for {} domain(s) '
            'with statuses: {}...'.format(field, statuses, qs.count()))
        for domain in qs:
            validate_sending_domain_field.apply_async([domain.id, field])


@task
def validate_sending_domain_field(domain_id, field):
    domain = SendingDomain.objects.get(id=domain_id)
    previous_status = getattr(domain, '{}_status'.format(field))
    getattr(domain, 'validate_{}'.format(field))()
    if getattr(domain, '{}_status'.format(field)) != previous_status:
        setattr(domain, '{}_status_date'.format(field), timezone.now())
        domain.save()



import logging

from django.core.management.base import BaseCommand

from ...models import Mail

log = logging.getLogger(__name__)


class Command(BaseCommand):
    help = 'List message statuses based on X-CM-Message-Id'

    def add_arguments(self, parser):
        parser.add_argument('identifier', nargs='+', type=str)
        parser.add_argument(
            '--with-headers',
            dest='with_headers',
            action='store_true')

    def handle(self, *args, **options):
        for mail in Mail.objects.filter(
                identifier__in=options['identifier']):
            self.stdout.write(self.style.SUCCESS(
                '* {} (pk:{}) (author:{})'.format(
                    mail.identifier, mail.pk, mail.author)))
            if options['with_headers']:
                print('├────────────── Headers ──────────────')
                for k, v in mail.headers.items():
                    print('│ {}: {}'.format(k, v))
            print('├────────────── Statuses ──────────────')
            for status in mail.mailstatus_set.all().order_by('creation_date'):
                print('├─ {}: {} by {} ({}) to {}'.format(
                    status.creation_date,
                    status.status,
                    status.source_ip,
                    status.source_hostname or '<missing>',
                    status.destination_domain or '<missing>'))
            print('└──────────────────────────────────────')



from rest_framework import serializers

from ...models import Category
from ...utils.serializers import HALLinksField


class CategorySerializer(serializers.HyperlinkedModelSerializer):
    _links = HALLinksField(
        nested_endpoints=[
            'opt_outs', 'stats', 'messages_stats'],
        view_name='core:category-detail')
    batches = serializers.HyperlinkedRelatedField(
        read_only=True, view_name='transactional:batch-detail',
        many=True, source='mailbatches')
    messages = serializers.HyperlinkedRelatedField(
        read_only=True, view_name='campaigns:message-detail', many=True)

    class Meta:
        model = Category
        fields = ('url', 'name', 'messages', 'batches', '_links')
        extra_kwargs = {'url': {'view_name': 'core:category-detail'}}



from django.conf import settings
from django.conf.urls import url
from django.conf.urls import include

from munch.urls import urlpatterns

from . import views

urlpatterns += [
    url(r'^_/contacts/', include([
        url(r'^confirm/(?P<uuid>.+)/?$', views.confirmation,
            name='confirmation'),
        url(r'^list/(?P<uuid>.+)/?$', views.subscription, name='subscription'),
    ]))
]

if settings.DEBUG:
    urlpatterns.insert(
        0, url(r'^test-form/(?P<uuid>.+)/?$', views.test_form, name='test'))



import email

from django.conf import settings
from django.db import transaction
from slimta.policy import QueuePolicy
from slimta.relay import PermanentRelayError

from munch.core.mail.models import RawMail
from munch.core.mail.utils import get_app_url
from munch.core.mail.utils import extract_domain
from munch.apps.campaigns.models import Category
from munch.apps.tracking.utils import get_msg_links
from munch.apps.optouts.contentfilters import set_unsubscribe_url
from munch.apps.tracking.contentfilters import add_tracking_image
from munch.apps.tracking.contentfilters import rewrite_html_links

from ...models import Mail
from ...models import MailBatch
from ...models import MailStatus


class Store(QueuePolicy):
    """
    Should occur after a n eventual return-path rewrite,
    as data is return-path-indexed.
    """

    @transaction.atomic()
    def apply(self, envelope):
        identifier = envelope.headers.get(
            settings.TRANSACTIONAL['X_MESSAGE_ID_HEADER'])
        if not identifier or not envelope.client.get('auth'):
            # FIXME: log exact error
            raise PermanentRelayError('Internal application error')

        # Process mail content
        munchers_kwargs = {
            'mail_identifier': identifier,
            'app_url': get_app_url(
                domain=envelope.sending_domain,
                organization=envelope.organization),
        }

        msg_links = {}
        mail_kwargs = {}
        track_open = envelope.headers.get(
            settings.TRANSACTIONAL.get('X_MAIL_TRACK_OPEN_HEADER', None))
        track_open_done = False
        if track_open:
            munchers_kwargs.update({'track_open': True})
            mail_kwargs.update({'track_open': True})

        track_clicks = envelope.headers.get(
            settings.TRANSACTIONAL.get('X_MAIL_TRACK_CLICKS_HEADER', None))
        if track_clicks:
            mail_kwargs.update({'track_clicks': True})
            munchers_kwargs.update({'track_clicks': True})

        add_unsubscribe = envelope.headers.get(
            settings.TRANSACTIONAL.get('X_MAIL_UNSUBSCRIBE_HEADER', None))
        if add_unsubscribe:
            munchers_kwargs.update({
                'unsubscribe_url': Mail.unsubscribe_url(
                    identifier, envelope.user, envelope.sending_domain)})

        # Retrieve full html to extract links
        html = ''
        message = email.message_from_bytes(b'\n'.join(envelope.flatten()))
        for part in message.walk():
            if part.get_content_type() == 'text/html':
                html += part.get_payload()

        msg_links = get_msg_links(html)
        mail_kwargs.update({'msg_links': msg_links})
        munchers_kwargs.update({'links_map': msg_links})

        # Walk throught every parts to apply munchers on it
        for part in message.walk():
            if part.get_content_type() == 'text/html':
                html = part.get_payload()
                if track_open and not track_open_done:
                    html = add_tracking_image(html, **munchers_kwargs)
                    track_open_done = True
                if track_clicks:
                    html = rewrite_html_links(html, **munchers_kwargs)
                part.set_payload(html)

            content = part.get_payload()

            if add_unsubscribe:
                content = set_unsubscribe_url(content, **munchers_kwargs)

            part.set_payload(content)

        envelope.parse_msg(message)

        batch = envelope.headers.get(
            settings.TRANSACTIONAL.get('X_MAIL_BATCH_HEADER', None))
        if batch:
            batch, created = MailBatch.objects.get_or_create(
                name=batch, author=envelope.user,
                defaults={'msg_links': msg_links})
            if not created:
                batch.msg_links = msg_links
                batch.save()
            category = envelope.headers.get(
                settings.TRANSACTIONAL.get(
                    'X_MAIL_BATCH_CATEGORY_HEADER', None))
            if category:
                category, _ = Category.objects.get_or_create(
                    author=envelope.user, name=category)
                batch.category = category
                batch.save()

        raw_mail, _ = RawMail.objects.get_or_create(content=envelope.message)
        mail = Mail.objects.create(
            author=envelope.user,
            batch=batch,
            identifier=envelope.headers.get(
                settings.TRANSACTIONAL['X_MESSAGE_ID_HEADER']),
            headers={k: v.encode(
                'utf-8', 'surrogateescape').decode('utf-8') for k, v in
                envelope.headers.raw_items()},
            message=raw_mail, sender=envelope.sender,
            recipient=envelope.recipients[0],
            **mail_kwargs)
        MailStatus.objects.create(
            mail=mail,
            destination_domain=extract_domain(envelope.recipients[0]))



from unittest import TestCase
from unittest.mock import patch
from unittest.mock import MagicMock
from urllib.parse import urlparse

import dns
from django.conf import settings

from munch.core.mail.utils import get_app_url
from munch.apps.domains.fields import DomainCheckField
from munch.apps.users.tests.factories import OrganizationFactory

from ..models import SendingDomain
from .factories import SendingDomainFactory


def DNSResultMock(v):
    """ Mocks DNS, single valued response

    DNSResultMock(42)
    DNSResultMock[1].to_text()
    """
    m = MagicMock()
    sub_mock = MagicMock()
    sub_mock.to_text.return_value = v
    m.__getitem__.return_value = sub_mock
    m.__iter__.return_value = [sub_mock]
    return m


def DNSEmptyEntryMock():
    m = MagicMock()
    m.__getitem__.side_effect = IndexError()
    return m


def mock_resolver(queries):
    mock_queries = queries

    def mock_resolve(name, field):
        try:
            res = mock_queries[name][field]
        except KeyError:
            raise dns.exception.DNSException
        else:
            if getattr(res, 'side_effect', None):
                return res()  # raise exception if needed
            else:
                return res
    m = MagicMock()
    m.side_effect = mock_resolve
    return m

APP_DOMAIN_CNAME = 'c.munch.local'
ok_app_domain = 'app.good.example.com'
bad_app_domain = 'app.bad.example.com'
notconfigured_app_domain = 'app.notconfigured.example.com'
inexistant_app_domain = 'donotexist.example.com'

mock_queries = {
    ok_app_domain: {
        'CNAME': DNSResultMock('{}.'.format(APP_DOMAIN_CNAME))},
    bad_app_domain: {'CNAME': DNSResultMock(
        'another.domain.com.')},
    notconfigured_app_domain: {'CNAME': DNSEmptyEntryMock()}}


class SendingDomainAppDomainTestCase(TestCase):
    def setUp(self):
        self.resolver = mock_resolver(mock_queries)
        settings.USERS['ORGANIZATION_APP_DOMAIN_CNAME'] = APP_DOMAIN_CNAME

    def test_app_domain_validate_ok(self):
        with patch('dns.resolver.query', self.resolver):
            domain = SendingDomainFactory(app_domain=ok_app_domain)
            domain.validate_app_domain()
            self.assertEqual(domain.app_domain_status, DomainCheckField.OK)

    def test_app_domain_validate_bad(self):
        with patch('dns.resolver.query', self.resolver):
            domain = SendingDomainFactory(app_domain=bad_app_domain)
            domain.validate_app_domain()
            self.assertEqual(
                domain.app_domain_status, DomainCheckField.BADLY_CONFIGURED)

    def test_app_domain_validate_notconfigured(self):
        with patch('dns.resolver.query', self.resolver):
            domain = SendingDomainFactory(app_domain=notconfigured_app_domain)
            domain.validate_app_domain()
            self.assertEqual(
                domain.app_domain_status, DomainCheckField.NOT_CONFIGURED)

    def test_app_domain_validate_inexistant(self):
        with patch('dns.resolver.query', self.resolver):
            domain = SendingDomainFactory(app_domain=inexistant_app_domain)
            domain.validate_app_domain()
            self.assertEqual(
                domain.app_domain_status, DomainCheckField.NOT_CONFIGURED)

    def test_get_app_url_default(self):
        org = OrganizationFactory()
        domain = SendingDomainFactory(organization=org)
        url = get_app_url(domain=domain, organization=org)
        self.assertEqual(url, settings.APPLICATION_URL)

    def test_get_app_url_good(self):
        with patch('dns.resolver.query', self.resolver):
            org = OrganizationFactory()
            domain = SendingDomainFactory(
                app_domain=ok_app_domain, organization=org)

            # retrieve object from DB to get latest app_domain_status
            domain = SendingDomain.objects.get(pk=domain.pk)

            url = get_app_url(domain=domain, organization=org)
            def_u = urlparse(settings.APPLICATION_URL.strip('/'))
            app_u = '{}://{}'.format(def_u.scheme, ok_app_domain)
            if def_u.port:
                app_u += ':{}'.format(def_u.port)
            self.assertEqual(url, app_u)

    def test_get_app_url_bad(self):
        with patch('dns.resolver.query', self.resolver):
            org = OrganizationFactory()
            domain = SendingDomainFactory(
                app_domain=bad_app_domain, organization=org)
            url = get_app_url(domain=domain, organization=org)
            self.assertEqual(url, settings.APPLICATION_URL)



#!/usr/bin/env python3
import os

from setuptools import setup
from setuptools import find_packages

base_dir = os.path.dirname(__file__)

about = {}
with open(os.path.join(base_dir, "src", "munch", "__about__.py")) as f:
    exec(f.read(), about)


with open(os.path.join(base_dir, "README.rst")) as f:
    long_description = f.read()

dependency_links = [
    # TODO: Remove dependency link when munch-mailsend will be published
   'git+https://github.com/crunchmail/munch-mailsend.git@master#egg=munch-mailsend-0.1.0']  # noqa


requires = [
    # Project
    'celery==3.1.23',
    'chardet==2.3.0',
    'clamd==1.0.2',
    'click==6.6',
    'cssutils==1.0.1',
    'Django==1.10.1',
    'django-filter==0.13.0',
    'django-fsm==2.4.0',
    'django-humanize==0.1.2',
    'djangorestframework==3.4.1',
    'djangorestframework-jwt==1.8.0',
    'djangorestframework-csv==1.4.1',
    'djangorestframework-bulk==0.2.1',
    'dnspython3==1.12.0',
    'html2text==2016.9.19',
    'Jinja2==2.8',
    'Markdown==2.6.6',
    'msgpack-python==0.4.8',
    'Pillow==3.3.0',
    'premailer==3.0.1',
    'Pygments==2.1.3',
    'python-magic>=0.4.10',
    'python-slimta==4.0.0',
    'requests==2.10.0',
    # munch-contacts
    'django-ipware==1.1.5',
    # Database
    'psycopg2==2.6.2',
    'psycogreen==1.0',
    # Cache
    'django-redis==4.4.4',
    # Fixtures
    'PyYAML==3.11',
    # Email backend
    'munch-mailsend==0.1.0',
]


setup(
    name=about["__title__"],
    version=about["__version__"],
    package_dir={'': 'src'},
    packages=find_packages('src'),
    long_description=long_description,
    include_package_data=True,
    description=about["__summary__"],
    author=about["__author__"],
    author_email=about["__email__"],
    url=about["__uri__"],
    install_requires=requires,
    dependency_links=dependency_links,
    license=about["__license__"],
    entry_points={
        'console_scripts': [
            'munch = munch.runner:main']
    },
    extras_require={
        'dev': [
            'bumpversion==0.5.3',
            'docker-compose==1.8.0',
            'mkdocs==0.16.0',
            'mkdocs-bootswatch==0.4.0'],
        'tests': [
            'flake8==3.2.0',
            'libfaketime==0.4.2',
            'factory-boy==2.7.0',
            'fake-factory==0.5.10'],
    },
    classifiers=[
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',  # noqa
    ]
)



from django.db import models
from django.core.validators import validate_email
from django.utils.translation import ugettext_lazy as _

from munch.apps.campaigns.models import Mail


class AbuseNotification(models.Model):
    mail = models.ForeignKey(Mail, verbose_name=_('mail'))
    date = models.DateTimeField(auto_now=True, verbose_name=_('date'))
    contact_name = models.CharField(
        _('contact name'), max_length=100, blank=True)
    contact_email = models.CharField(
        _('contact email'), validators=[validate_email],
        max_length=50, blank=True)
    comments = models.TextField(_('comments'))

    class Meta:
        verbose_name = _("abuse report")
        verbose_name_plural = _("abuse reports")

    def __str__(self):
        return _('Abuse report by {}').format(self.mail.recipient)



from django.test import TestCase
from django.test.utils import override_settings
from django.core.exceptions import ValidationError

from .factories import MailFactory
from .factories import MessageFactory
from munch.apps.users.tests.factories import UserFactory


class MailValidatorTest(TestCase):
    def setUp(self):
        self.user = UserFactory()

    def test_ok(self):
        message = MessageFactory(author=self.user)
        with override_settings(BYPASS_DNS_CHECKS=False):
            mail = MailFactory(message=message)
            mail.full_clean()

    def test_not_an_email(self):
        message = MessageFactory(author=self.user)
        with self.assertRaisesRegexp(ValidationError, 'a valid email address'):
            with override_settings(BYPASS_DNS_CHECKS=False):
                mail = MailFactory(
                    recipient='eitsuanetisanetsunaetsurn', message=message)
                mail.full_clean()

    def test_not_an_email_bis(self):
        message = MessageFactory(author=self.user)
        with self.assertRaisesRegexp(ValidationError, 'a valid email address'):
            with override_settings(BYPASS_DNS_CHECKS=False):
                mail = MailFactory(message=message, recipient='foo@bar')
                mail.full_clean()

    def test_not_a_valid_domain(self):
        message = MessageFactory(author=self.user)
        # example.com has no MX record
        with self.assertRaisesRegexp(ValidationError, '^((?!invalide).)*$'):
            with override_settings(BYPASS_DNS_CHECKS=False):
                message.sender_email = 'foo@example.com'
                message.save()

        mail = MailFactory(
            message=message, recipient='test@example.com')
        mail.full_clean()
        message.save()



import collections

from django.apps import apps
from django.conf import settings
from rest_framework.reverse import reverse
from rest_framework.response import Response
from rest_framework.decorators import api_view


@api_view(('GET',))
def api_root(request, format=None):
    if request.version == 'v1':
        data = {
            # Core
            'categories': reverse(
                'v1:core:category-list', request=request),
            # Campaigns
            'messages': reverse('v1:campaigns:message-list', request=request),
            'recipients': reverse(
                'v1:campaigns:recipient-list', request=request),
            'bounces': reverse('v1:campaigns:bounce-list', request=request),
            'attachments': reverse(
                'v1:campaigns:messageattachment-list', request=request),
            # Optouts
            'opt-outs': reverse('v1:optouts:opt-outs-list', request=request),
            # Domains
            'domains': reverse(
                'v1:domains:sendingdomain-list', request=request),
            # Users
            'applications/api': reverse(
                'v1:users:applications-api-list', request=request),
            'applications/smtp': reverse(
                'v1:users:applications-smtp-list', request=request),
            'me': reverse('v1:users:me', request=request),
            'organizations': reverse(
                'v1:users:organization-list', request=request),
            'users': reverse('v1:users:munchuser-list', request=request),
            # Upload store
            'images': reverse('v1:upload-store:image-create', request=request),
            # Transactional
            'transactional/mails': reverse(
                'v1:transactional:mail-list', request=request),
            'transactional/batches': reverse(
                'v1:transactional:batch-list', request=request)}
    elif request.version == 'v2':
        data = {
            # Core
            'categories': reverse(
                'v2:core:category-list', request=request),
            # Campaigns
            'messages': reverse('v2:campaigns:message-list', request=request),
            'recipients': reverse(
                'v2:campaigns:recipient-list', request=request),
            'bounces': reverse('v2:campaigns:bounce-list', request=request),
            'attachments': reverse(
                'v2:campaigns:messageattachment-list', request=request),
            # Optouts
            'opt-outs': reverse('v2:optouts:opt-outs-list', request=request),
            # Domains
            'domains': reverse(
                'v2:domains:sendingdomain-list', request=request),
            # Users
            'applications/api': reverse(
                'v2:users:applications-api-list', request=request),
            'applications/smtp': reverse(
                'v2:users:applications-smtp-list', request=request),
            'me': reverse('v2:users:me', request=request),
            'organizations': reverse(
                'v2:users:organization-list', request=request),
            'users': reverse('v2:users:munchuser-list', request=request),
            # Upload store
            'images': reverse('v2:upload-store:image-create', request=request),
            # Transactional
            'transactional/mails': reverse(
                'v1:transactional:mail-list', request=request),
            'transactional/batches': reverse(
                'v1:transactional:batch-list', request=request)}
    data.update({
        # Upload store
        'images': reverse('upload-store:image-create', request=request)})

    # Import external apps
    for name, config in apps.app_configs.items():
        if hasattr(config, 'is_munch_app') and config.is_munch_app:
            try:
                api_module = config.module.api
                try:
                    api_module = getattr(config.module.api, request.version)
                except AttributeError:
                    pass
                urls = api_module.get_api_root_urls(request)
                data.update(urls)
            except AttributeError:
                pass
            except:
                if settings.DEBUG:
                    raise

    return Response(collections.OrderedDict(
        sorted(data.items(), key=lambda t: t[0])))



import os

from django.core.files.storage import FileSystemStorage


class LocalFileSystemStorage(FileSystemStorage):

    def __init__(self, base_url, *args, **kwargs):
        super().__init__(base_url=base_url, *args, **kwargs)

    def get_available_name(self, name, max_length=None):
        # We assume that our name was properly set to a
        # unique (per organization) hash by the model
        return name

    def _save(self, name, content):
        # TODO: use content.kind set by upstream save() to force
        # Content-Disposition: attachment on type 'file'
        # using original file name (content.name)
        full_path = self.path(name)
        if os.path.exists(full_path):
            # If we're saving a duplicate, overwrite the file.
            # Needed because upstream _save() will request a new file name
            # if it already exists
            self.delete(name)

        return super()._save(name, content)



import math
from collections import OrderedDict

from rest_framework.response import Response
from rest_framework.pagination import PageNumberPagination


class CountedPageNumberPagination(PageNumberPagination):
    """ Simple page pagination indicating the total numbre of pages.
    """
    UNPAGINATED_HARD_LIMIT = 2000000

    page_size = 50
    max_page_size = 1000
    page_size_query_param = 'page_size'

    def get_page_count(self):
        return math.ceil(
            self.page.paginator.count / self.get_page_size(self.request))

    def get_page_size(self, request):
        skip_pagination = getattr(
            request.accepted_renderer, 'skip_pagination', False)
        if skip_pagination:
            return self.UNPAGINATED_HARD_LIMIT
        else:
            return super().get_page_size(request)

    def get_paginated_response(self, data):
        return Response(OrderedDict([
            ('count', self.page.paginator.count),
            ('next', self.get_next_link()),
            ('previous', self.get_previous_link()),
            ('page_count', self.get_page_count()),
            ('results', data)]))



from importlib import import_module
from django.conf import settings

from .celery import app  # noqa
from .celery import munch_tasks_router  # noqa
from .celery import add_queues as add_core_queues
from .celery import register_tasks as register_core_tasks

from .utils import monkey_patch_slimta_exception

default_app_config = 'munch.core.apps.CoreApp'


add_core_queues()
register_core_tasks()


# Register apps queues, tasks and workers
for _app in settings.INSTALLED_APPS:
    try:
        app_celery = import_module('{}.celery'.format(_app))
        try:
            app_celery.add_queues()
        except AttributeError:
            pass
        try:
            app_celery.register_tasks()
        except AttributeError:
            pass
    except ImportError:
        pass


def register_extended_msgpack():
    # Register extended msgpack decoder
    import datetime
    import pytz
    from slimta.smtp.reply import Reply

    import msgpack
    from kombu.serialization import register

    def msgpack_decode(obj):
        def object_hook(obj):
            if '__datetime__' in obj:
                obj = datetime.datetime.strptime(
                    obj['as_str'], "%Y%m%dT%H:%M:%S.%f").replace(
                        tzinfo=pytz.utc)
            elif '__slimta.Reply__' in obj:
                obj = Reply(
                    code=obj['as_dict'].get('code'),
                    message=obj['as_dict'].get('message'),
                    command=obj['as_dict'].get('command'),
                    address=obj['as_dict'].get('address'))
            return obj
        return msgpack.unpackb(obj, object_hook=object_hook, encoding='utf-8')

    def msgpack_encode(obj):
        def hook(obj):
            if isinstance(obj, datetime.datetime):
                return {
                    '__datetime__': True,
                    'as_str': obj.strftime("%Y%m%dT%H:%M:%S.%f")}
            elif isinstance(obj, Reply):
                return {
                    '__slimta.Reply__': True,
                    'as_dict': {
                        'code': obj.code,
                        'address': obj.address,
                        'message': obj.message,
                        'command': obj.command,
                    }
                }
            return obj
        return msgpack.packb(obj, default=hook, use_bin_type=True)

    register(
        'extended-msgpack', msgpack_encode, msgpack_decode,
        content_type='application/x-extended-msgpack',
        content_encoding='binary')


register_extended_msgpack()
monkey_patch_slimta_exception()



from ...utils.permissions import MunchResourcePermission


class CategoryPermission(MunchResourcePermission):
    app_name = 'core'
    model_name = 'category'

    def has_object_permission(self, request, view, obj):
        if request.method == 'DELETE':
            if obj.messages.exists():
                return False

        return super().has_object_permission(request, view, obj)



import logging

from django.conf import settings
from django.db.models import Prefetch
from slimta.policy import QueuePolicy
from slimta.queue import QueueError
from slimta.smtp.reply import Reply

from munch.apps.optouts.models import OptOut
from munch.apps.users.models import MunchUser
from munch.apps.users.models import Organization
from munch.apps.users.models import SmtpApplication

logger = logging.getLogger(__name__)


class Check(QueuePolicy):
    def apply(self, envelope):
        error = QueueError()
        error.reply = Reply(
            code='521',
            message='5.7.1 Email to this recipient will not be delivered')

        if OptOut.objects.filter(
                origin=OptOut.BY_BOUNCE,
                address=envelope.recipients[0]).exists():
            raise error

        # Sorry about that, it's just a big query set with prefetch related
        user = SmtpApplication.objects.prefetch_related(
            Prefetch(
                'author',
                MunchUser.objects.prefetch_related(
                    Prefetch('organization', Organization.objects.all().only(
                        'id'))).all().only('organization'))).only(
                            'author').get(username=envelope.client.get(
                                'auth')[0]).author

        # Attach theses attributs to envelope to avoid re-requesting it later
        envelope.user = user
        envelope.organization = user.organization

        category = envelope.headers.get(
            settings.TRANSACTIONAL.get(
                'X_MAIL_BATCH_CATEGORY_HEADER', None))
        if category:
            if OptOut.objects.exclude(
                    origin=OptOut.BY_BOUNCE).filter(
                        author__organization=envelope.organization,
                        category__name=category).exists():
                raise error
        else:
            if OptOut.objects.exclude(
                    origin=OptOut.BY_BOUNCE).filter(
                        author__organization=envelope.organization).exists():
                raise error



import unittest
from datetime import datetime
from datetime import timedelta

from django.core.cache import cache
from django.test import TestCase
from django.utils.timezone import utc
from django.test.utils import override_settings
from libfaketime import fake_time

from munch.core.utils.tests import temporary_settings
from munch.apps.users.tests.factories import UserFactory

from ..tasks import handle_dsn
from ..tasks import handle_fbl
from ..tasks import handle_mail_optout
from ..models import Mail
from ..models import OptOut
from ..models import Message
from ..models import MailStatus

from .factories import MailFactory
from .factories import MessageFactory
from .factories import MailStatusFactory


@override_settings(
    SKIP_SPAM_CHECK=False, SPAMD_SERVICE_NAME='dne.example.com',
    SPAMD_HOST='dne.example.com')
class DSNQTests(TestCase):
    """
    SPAMD_SERVICE_NAME is set to something wrong
    to be sure it never gets called
    """
    def setUp(self):
        cache.clear()
        # To start testing DSN status, we have first to simulate that the mail
        # has been passed to infrastructure.
        self.user = UserFactory()
        self.message = MessageFactory(author=self.user)

        self.mail_1 = MailFactory(message=self.message)
        self.mail_2 = MailFactory(message=self.message)

        details = (
            (self.mail_1, MailStatus.QUEUED, datetime(
                2014, 8, 5, 22, tzinfo=utc) + timedelta(minutes=5)),
            (self.mail_1, MailStatus.SENDING, datetime(
                2014, 8, 6, 22, tzinfo=utc)),
            (self.mail_2, MailStatus.QUEUED, datetime(
                2014, 8, 5, 22, tzinfo=utc) + timedelta(minutes=5)),
            (self.mail_2, MailStatus.SENDING, datetime(
                2014, 8, 6, 22, tzinfo=utc)))

        for mail, status, date in details:
            MailStatusFactory(
                mail=mail, status=status, creation_date=date, raw_msg=status)

        self.hardbounce_message = (
            "smtp; 550 5.1.1 <nonexist@example.com>: Recipient address"
            "rejected: User unknown in virtual mailbox table")

    def test_mail_msg_ok(self):
        with fake_time('Fri,  5 Aug 2014 23:35:52 +0700 (WIT)'):
            handle_dsn(
                'To: return-{}@test.munch.example.com'.format(
                    self.mail_1.identifier),
                {
                    'Final-Recipient': self.mail_1.recipient,
                    'Diagnostic-Code': 'Delivered',
                    'Status': '2.0.0',
                    'Arrival-Date': 'Fri,  5 Aug 2014 23:35:50 +0700 (WIT)'})
        self.mail_1 = Mail.objects.get(pk=self.mail_1.pk)
        self.assertEqual(self.mail_1.curstatus, MailStatus.DELIVERED)

        last_status = self.mail_1.statuses.last()
        self.assertEqual(last_status.status, MailStatus.DELIVERED)
        self.assertEqual(
            last_status.creation_date, datetime(
                2014, 8, 5, 21, 35, 52,
                tzinfo=last_status.creation_date.tzinfo))
        self.assertEqual(last_status.mail, self.mail_1)
        self.assertEqual(last_status.raw_msg, 'Delivered')
        self.assertEqual(last_status.status_code, '2.0.0')

    def test_mail_msg_to_protected(self):
        """ Check that we extract the To from optional surounding "<" ">"
        """
        with fake_time('Fri,  5 Aug 2014 23:35:52 +0700 (WIT)'):
            handle_dsn(
                'To: A A <return-{}@test.munch.example.com>'.format(
                    self.mail_1.identifier),
                {
                    'Final-Recipient': self.mail_1.recipient,
                    'Diagnostic-Code': 'Delivered',
                    'Status': '2.0.0',
                    'Arrival-Date': 'Fri,  5 Aug 2014 23:35:50 +0700 (WIT)'})
        self.mail_1 = Mail.objects.get(pk=self.mail_1.pk)
        self.assertEqual(self.mail_1.curstatus, MailStatus.DELIVERED)

    def test_mail_msg_hard(self):
        handle_dsn(
            'To: return-{}@test.munch.example.com'.format(self.mail_1.identifier),
            {
                'Final-Recipient': self.mail_1.recipient,
                'Diagnostic-Code': self.hardbounce_message,
                'Status': '5.1.1',
                'Arrival-Date': 'Fri,  5 Aug 2014 23:35:50 +0700 (WIT)'})
        self.mail_1 = Mail.objects.get(pk=self.mail_1.pk)
        self.assertEqual(self.mail_1.curstatus, MailStatus.BOUNCED)

        last_status = self.mail_1.statuses.last()
        self.assertEqual(last_status.status, MailStatus.BOUNCED)
        self.assertEqual(
            last_status.raw_msg, self.hardbounce_message.replace('\n', ' '))

    def test_mail_msg_soft(self):
        handle_dsn(
            'To: return-{}@test.munch.example.com'.format(self.mail_1.identifier),
            {
                'Final-Recipient': self.mail_1.recipient,
                'Action': 'Failed',
                'Diagnostic-Code': 'SoftBounce',
                'Status': '4.1.1',
                'Arrival-Date': 'Fri,  5 Aug 2014 23:35:50 +0700 (WIT)'})
        self.mail_1 = Mail.objects.get(pk=self.mail_1.pk)
        self.assertEqual(self.mail_1.curstatus, MailStatus.DROPPED)

        last_status = self.mail_1.statuses.last()
        self.assertEqual(last_status.status, MailStatus.DROPPED)
        self.assertEqual(last_status.raw_msg, 'SoftBounce')

    def test_mail_msg_soft_notify(self):
        """ Quite rare setup where the MTA nofifies the envelope-from

        when its message is being requeued.
        """
        handle_dsn(
            'To: return-{}@test.munch.example.com'.format(self.mail_1.identifier),
            {
                'Final-Recipient': self.mail_1.recipient,
                'Action': 'delayed',
                'Diagnostic-Code': 'SoftBounce',
                'Status': '4.1.1',
                'Arrival-Date': 'Fri,  5 Aug 2014 23:35:50 +0700 (WIT)'})
        self.mail_1 = Mail.objects.get(pk=self.mail_1.pk)
        self.assertEqual(self.mail_1.curstatus, MailStatus.DELAYED)

        last_status = self.mail_1.statuses.last()
        self.assertEqual(last_status.status, MailStatus.DELAYED)
        self.assertEqual(last_status.raw_msg, 'SoftBounce')

    def test_last_mail_change_status(self):
        self.message.status = 'sending'
        self.message.save()

        handle_dsn(
            'To: return-{}@test.munch.example.com'.format(self.mail_1.identifier),
            {
                'Final-Recipient': self.mail_1.recipient,
                'Diagnostic-Code': 'Delivered',
                'Status': '2.0.0',
                'Arrival-Date': 'Fri,  5 Aug 2014 23:35:50 +0700 (WIT)'})
        self.assertEqual(self.message.status, Message.SENDING)
        handle_dsn(
            'To: return-{}@test.munch.example.com'.format(self.mail_2.identifier),
            {
                'Final-Recipient': self.mail_2.recipient,
                'Diagnostic-Code': self.hardbounce_message,
                'Status': '5.1.1',
                'Arrival-Date': 'Fri,  5 Aug 2014 23:35:50 +0700 (WIT)'})
        self.message = Message.objects.get(pk=self.message.pk)
        self.assertEqual(self.message.status, Message.SENT)
        self.assertIsNotNone(self.message.completion_date)

    def test_bounce_after_delivered(self):
        """
        Test that we can become « softbounced » after having been « delivered »

        That's strange, but may happen if an intermediary server ack the
        message and the final one bounce it with backscatter.
        """
        handle_dsn(
            'To: return-{}@test.munch.example.com'.format(self.mail_1.identifier),
            {
                'Final-Recipient': self.mail_1.recipient,
                'Diagnostic-Code': 'Delivered',
                'Status': '2.0.0',
                'Arrival-Date': 'Fri,  5 Aug 2014 23:35:50 +0700 (WIT)'})
        handle_dsn(
            'To: return-{}@test.munch.example.com'.format(self.mail_1.identifier),
            {
                'Final-Recipient': self.mail_1.recipient,
                'Diagnostic-Code': self.hardbounce_message,
                'Status': '5.1.1',
                'Arrival-Date': 'Fri,  5 Aug 2014 23:50:50 +0700 (WIT)'})
        self.mail_1 = Mail.objects.get(pk=self.mail_1.pk)
        self.assertEqual(self.mail_1.curstatus, MailStatus.BOUNCED)

        last_status = self.mail_1.statuses.last()
        self.assertEqual(last_status.status, MailStatus.BOUNCED)
        self.assertEqual(
            last_status.raw_msg, self.hardbounce_message.replace('\n', ' '))

    @override_settings(BOUNCE_POLICY=[(['5.'], 3, 365)])
    def test_several_bounce_imply_optout(self):
        """ Make an optout after a certain of ammount of bounce is recvd"""
        user = UserFactory()

        def mk_message(i):
            with self.settings(SKIP_SPAM_CHECK=True):
                message = MessageFactory(author=user)
                message.status = Message.MSG_OK
                message.save()
                message.status = Message.SENDING
                message.save()
                mail = MailFactory(message=message, recipient='i@example.com')

                # Simulate it has been passed to infrastructure
                MailStatusFactory(
                    mail=mail, status=MailStatus.QUEUED,
                    raw_msg='sent by local MTA',
                    creation_date=datetime(2014, 8, 4, tzinfo=utc))
                MailStatusFactory(
                    mail=mail, status=MailStatus.SENDING,
                    raw_msg='sent by local MTA',
                    creation_date=datetime(2014, 8, 4, tzinfo=utc))

            handle_dsn(
                'To: {}'.format(mail.envelope_from),
                {
                    'Final-Recipient': 'i@example.com',
                    'Diagnostic-Code': self.hardbounce_message,
                    'Status': '5.1.1',
                    'Arrival-Date': 'Fri,  5 Aug 2014 23:35:50 +0700 (WIT)'})

        optout = OptOut.objects.filter(address='i@example.com')

        mk_message(1)
        self.assertFalse(optout.exists())
        mk_message(2)
        self.assertFalse(optout.exists())
        mk_message(3)
        self.assertTrue(optout.exists())

    def test_fbl_optout(self):
        main_headers = (
            'To: abuse@test.munch.example.com\n'
            'Date: Fri,  5 Aug 2014 23:35:50 +0700 (WIT)')
        orig_headers = {
            'Return-Path': 'return-{}@test.munch.example.com'.format(
                self.mail_1.identifier)}
        feedback_headers = {
            'Feedback-Type': 'abuse',
            'User-Agent': 'unittest',
            'Version': '0.1'}
        handle_fbl(main_headers, orig_headers, feedback_headers)

        optout = OptOut.objects.get(identifier=self.mail_1.identifier)
        d = datetime(
            2014, 8, 5, 16, 35, 50, tzinfo=optout.creation_date.tzinfo)

        self.assertEqual(optout.creation_date, d)
        self.assertEqual(optout.origin, OptOut.BY_FBL)

    def test_mail_optout(self):
        headers = {
            'To': 'unsubscribe-{}@test.munch.example.com'.format(
                self.mail_1.identifier),
            'Date': 'Fri,  5 Aug 2014 23:35:50 +0700 (WIT)'}
        handle_mail_optout(headers)

        optout = OptOut.objects.get(identifier=self.mail_1.identifier)
        d = datetime(
            2014, 8, 5, 16, 35, 50, tzinfo=optout.creation_date.tzinfo)
        self.assertEqual(optout.creation_date, d)
        self.assertEqual(optout.origin, OptOut.BY_MAIL)

    def test_mail_multiple_optout(self):
        """ Try to send multiple optout. Second one must be ignored. """
        headers = {
            'To': 'unsubscribe-{}@test.munch.example.com'.format(
                self.mail_1.identifier),
            'Date': 'Fri,  5 Aug 2014 23:35:50 +0700 (WIT)'}
        handle_mail_optout(headers)
        headers = {
            'To': 'unsubscribe-{}@test.munch.example.com'.format(
                self.mail_1.identifier),
            'Date': 'Fri,  5 Aug 2014 23:37:50 +0700 (WIT)'}

        handle_mail_optout(headers)
        optout = OptOut.objects.get(identifier=self.mail_1.identifier)
        d = datetime(
            2014, 8, 5, 16, 35, 50, tzinfo=optout.creation_date.tzinfo)
        self.assertEqual(optout.creation_date, d)
        self.assertEqual(optout.origin, OptOut.BY_MAIL)

    @unittest.skip('Missing test')
    def test_invalid_data_reject(self):
        """ Invalid data should trigger logging and rejecting
        """
        pass

    @unittest.skip('Missing test')
    def test_invalid_transition_reject(self):
        """ Invalid data should trigger logging and rejecting
        """
        pass



import os
import shutil
import logging

from django.test import Client
from django.test import TestCase
from django.conf import settings
from django.test.utils import override_settings

from .models import InlineImage
from .models import HostedImage
from .exceptions import TooBigMedia
from .exceptions import InvalidMimeType

from munch.apps.users.tests.factories import UserFactory
from munch.apps.campaigns.tests.factories import MessageFactory


logging.disable(logging.CRITICAL)


class TestHostedMedia(TestCase):
    """ Deletes the media dir once tests are completed

    WARNING: use that only if you are in tests, with an overloaded  MEDIA_ROOT
    Normally it's fine within munch, defined in munch/test_settings.py
    """
    def classSetUp(self):
        # mktemp will only be used for the 1st Case using this parent class
        if not os.path.exists(settings.MEDIA_ROOT):
            os.mkdir(settings.MEDIA_ROOT)

    def classTearDown(self):
        try:
            shutil.rmtree(settings.MEDIA_ROOT)
        except Exception:
            pass


class TestHostedImage(TestHostedMedia):
    def setUp(self):
        self.user = UserFactory()

    def test_image_toobig(self):
        class SmallHostedImage(HostedImage):
            # very small
            MAX_SIZE = 100
        img_url = 'http://www.oasiswork.fr/uploads/logo_oasiswork.jpg'

        with self.assertRaises(TooBigMedia):
            img = SmallHostedImage(
                img_url, organization=self.user.organization)
            img.store()

    def test_wrong_mimetype(self):
        html_document = 'data:image/gif;base64,PGh0bWw+PC9odG1sPg=='
        image = InlineImage(html_document, organization=self.user.organization)
        with self.assertRaises(InvalidMimeType):
            image.store()


class TestHostedImageClient(TestHostedMedia):
    def setUp(self):
        self.client = Client()
        self.user = UserFactory()

    def test_fetch_hosted_image_ok(self):
        img_url = 'http://www.oasiswork.fr/uploads/logo_oasiswork.jpg'
        img = HostedImage(img_url, organization=self.user.organization)
        stored_url = img.store()
        self.assertTrue(stored_url.startswith('http://munch.example.com'))

    def test_fetch_inline_image_ok(self):
        img_data = (
            "data:image/gif;base64,R0lGODlhDwAPAKECAAAAzMzM/////"
            "wAAACwAAAAADwAPAAACIISPeQHsrZ5ModrLlN48CXF8m2iQ3YmmKqVlRtW4ML"
            "wWACH+H09wdGltaXplZCBieSBVbGVhZCBTbWFydFNhdmVyIQAAOw==")
        img = InlineImage(img_data, organization=self.user.organization)
        stored_url = img.store()
        self.assertTrue(stored_url.startswith('http://munch.example.com'))


@override_settings(SECRET_KEY='123412341234')
class TestHostedMessage(TestCase):
    def setUp(self):
        self.user = UserFactory()
        self.message = MessageFactory(author=self.user)

    def test_hosted_mail_without_key_works(self):
        response = self.client.get('/archive/{}/'.format(
            self.message.identifier))
        self.assertEqual(response.status_code, 200)

    def test_hosted_mail_bad_uuid(self):
        response = self.client.get('/archive/1584498-148484/')
        self.assertEqual(response.status_code, 404)



import json

from rest_framework import renderers
from rest_framework_csv.renderers import CSVRenderer


class PlainTextRenderer(renderers.BaseRenderer):
    media_type = 'text/plain'
    format = 'txt'

    def render(self, data, media_type=None, renderer_context=None):
        return str(data).encode(self.charset)


class HTMLRenderer(renderers.BaseRenderer):
    media_type = 'text/html'
    format = 'html'

    def render(self, data, media_type=None, renderer_context=None):
        return str(data).encode(self.charset)


class PaginatedCSVRenderer (CSVRenderer):
    results_field = 'results'
    skip_pagination = True

    def render(self, data, media_type=None, renderer_context=None):
        if not isinstance(data, list):
            data = data.get(self.results_field, [])
        return super().render(data, media_type, renderer_context)

    def flatten_list(self, l):
        """ Renders the lists as a raw json-list

        Otherwise, the output number of columns is unpredictable
        """
        return {'': json.dumps(l)}



import re

# from https://mail.python.org/pipermail/tutor/2002-February/012481.html

urls = '(%s)' % '|'.join("http https ftp".split())
ltrs = r'\w'
gunk = r'/#~:.?+=&%@!\-'
punc = r'.:?\-'
any = "%s%s%s" % (ltrs, gunk, punc)

url = r"""
    (                             # begin \1 {
        %(urls)s    :             # need resource and a colon
        [%(any)s] +?              # followed by one or more
                                  #  of any valid character, but
                                  #  be conservative and take only
                                  #  what you need to....
    )                             # end   \1 }
    (?=                           # look-ahead non-consumptive assertion
            [%(punc)s]*           # either 0 or more punctuation
            [^%(any)s]            #  followed by a non-url char
        |                         # or else
            $                     #  then end of the string
    )
    """ % {'urls': urls,
           'any': any,
           'punc': punc}

url_re = re.compile(url, re.VERBOSE)
mkd_footnote_url_re = re.compile(
    r'(?P<fnmark>\[\d+\]\:)\s(?P<url>{})'.format(url), re.VERBOSE)



import os
import copy
import logging
from datetime import timedelta

from django.utils.log import DEFAULT_LOGGING
from django.utils.translation import ugettext_lazy as _

##########
# Django #
##########
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
SECRET_KEY = 'CHANGEME'
DEBUG = False
ALLOWED_HOSTS = []
INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django.contrib.admin',
    'django.contrib.postgres',
    # External libs
    'django_humanize',
    'rest_framework',
    # In-project apps
    'munch.core',
    'munch.apps.users',
    'munch.apps.abuse',
    'munch.apps.contacts',
    'munch.apps.campaigns',
    'munch.apps.domains',
    'munch.apps.optouts',
    'munch.apps.hosted',
    'munch.apps.spamcheck',
    'munch.apps.tracking',
    'munch.apps.transactional',
    'munch.apps.upload_store',
    # Mail backend
    'munch_mailsend',
]
MIDDLEWARE_CLASSES = [
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.locale.LocaleMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.auth.middleware.SessionAuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'django.middleware.security.SecurityMiddleware',
]
ROOT_URLCONF = 'munch.urls'
AUTH_USER_MODEL = 'users.MunchUser'
AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',  # noqa
        'OPTIONS': {
            'user_attributes': ['identifier', 'first_name', 'last_name'],
            'max_similarity': 0.7,
        }
    },
    {'NAME': 'munch.apps.users.validators.MinimumLengthValidator'},
    {'NAME': 'munch.apps.users.validators.MinimumDigitValidator'},
    {'NAME': 'munch.apps.users.validators.MinimumUppercaseValidator'},
]
WSGI_APPLICATION = 'munch.wsgi.application'
LANGUAGE_CODE = 'en-us'
LANGUAGES = [
    ('fr', _('French')),
    ('en', _('English')),
]
DEFAULT_CHARSET = 'utf-8'
LOCALE_PATHS = [os.path.join(BASE_DIR, 'locale')]
TIME_ZONE = 'Europe/Paris'
USE_I18N = True
USE_L10N = True
USE_TZ = True
TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [os.path.join(BASE_DIR, 'templates')],
        'APP_DIRS': True,
        'OPTIONS': {
            'debug': DEBUG,
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]
STATIC_URL = '/static/'
STATICFILES_DIRS = (
    os.path.join(BASE_DIR, 'assets'),
)
MEDIA_URL = '/media/'
MEDIA_ROOT = os.path.join(BASE_DIR, 'media')
AUTHENTICATION_BACKENDS = [
    'django.contrib.auth.backends.ModelBackend',
    'munch.apps.users.backends.SecretMunchUserBackend']
BROKER_URL = 'amqp://localhost:15642'

###########
# Logging #
###########
LOGGING = copy.deepcopy(DEFAULT_LOGGING)
LOGGING['formatters'] = {
    'simple': {
        'format':
            '[%(asctime)s: %(levelname)s/%(module)s.%(funcName)s] %(message)s'
    },
    'django.server': {
        '()': 'django.utils.log.ServerFormatter',
        'format': '[%(server_time)s] %(message)s',
    }
}
# Log every level. Note: need setting DEBUG=True
LOGGING['handlers']['console']['level'] = logging.NOTSET
LOGGING['handlers']['console']['filters'] = []
LOGGING['handlers']['console']['formatter'] = 'simple'
LOGGING['handlers']['django.server'] = {
    'level': 'INFO',
    'class': 'logging.StreamHandler',
    'formatter': 'django.server',
}

# Define custom log handlers for slimta.
# Define a common parent for all slimta children loggers. By default,  we only
# dispatch logs with an INFO level to handlers.
LOGGING['loggers']['slimta'] = {
    'handlers': ['console'],
    'level': logging.INFO,
}
# Define custom project loggers.
LOGGING['loggers']['munch'] = {
    'handlers': ['console'],
    'filters': [],
    'propagate': False,
    'level': logging.INFO,
}
LOGGING['loggers']['munch.errors'] = {
    'handlers': ['mail_admins'],
    'level': logging.ERROR,
    'propagate': False,
}
LOGGING['loggers']['django.server'] = {
    'handlers': ['django.server'],
    'level': 'INFO',
    'propagate': False,
}

##########
# Celery #
##########
CELERY_ACKS_LATE = True
CELERY_CREATE_MISSING_QUEUES = True
CELERY_DEFAULT_QUEUE = 'munch.default'
CELERY_DEFAULT_EXCHANGE = 'munch'
CELERY_DEFAULT_EXCHANGE_TYPE = 'topic'
CELERY_DEFAULT_ROUTING_KEY = 'munch.unknown'
CELERY_ACCEPT_CONTENT = ['extended-msgpack']
CELERY_TASK_SERIALIZER = 'extended-msgpack'
CELERY_RESULT_SERIALIZER = 'extended-msgpack'
CELERY_ROUTES = ('munch.core.celery.CeleryRouter',)

CELERYBEAT_SCHEDULE = {
    'purge_raw_mail': {
        'task': 'munch.core.mail.tasks.purge_raw_mail',
        'schedule': timedelta(hours=2),
    },
    'run_badly_configured_domains_validation': {
        'task': 'munch.apps.domains.tasks.run_domains_validation',
        'args': (['ko', 'bad', 'unknown'], ),
        'schedule': timedelta(minutes=15),
    },
    'run_well_configured_domains_validation': {
        'task': 'munch.apps.domains.tasks.run_domains_validation',
        'args': (['ok'], ),
        'schedule': timedelta(hours=12),
    },
}

##########
# Global #
##########
# Product name appear in email templates and optouts pages
PRODUCT_NAME = 'Munch'
# Default return-path to use (must be backmuncher MX)
RETURNPATH_DOMAIN = 'munch.example.com'
# URL base used for links generation (optouts, tracking, subscriptions, ...)
APPLICATION_URL = 'http://localhost:8000'
# URL base for internal pages (password reset, invitation validation, ...)
HELPDESK_URL = 'http://localhost:8000'

LOGIN_URL = 'rest_framework:login'
LOGIN_REDIRECT_URL = '/v1'

SERVICE_MSG_FROM_NAME = PRODUCT_NAME
SERVICE_MSG_FROM_EMAIL = 'no-reply@{}'.format(RETURNPATH_DOMAIN)
SERVICE_MSG_RETURN_PATH = SERVICE_MSG_FROM_EMAIL
NOTIFICATION_MSG_FROM_NAME = SERVICE_MSG_FROM_NAME
NOTIFICATION_MSG_FROM_EMAIL = SERVICE_MSG_FROM_EMAIL

PASSWORD_MIN_LENGTH = 6
PASSWORD_MIN_DIGIT = 1
PASSWORD_MIN_UPPER = 1

PASSWORD_RESET_TIMEOUT_DAYS = 1
INVITATION_TIMEOUT_DAYS = 15
ACCOUNT_ACTIVATION_TIMEOUT_DAYS = 30

EMAIL_BACKEND = 'munch_mailsend.backend.Backend'
MASS_EMAIL_BACKEND = 'munch_mailsend.backend.Backend'
# May be set True for debug/testing
BYPASS_DNS_CHECKS = False

GOOGLE_FBL_SENDER_ID = 'munch'

MSGID_DOMAIN = None

# Enable performance stats recording via Statsd
STATSD_ENABLED = False
STATSD_HOST = 'localhost'
STATSD_PORT = 8125
STATSD_PREFIX = None
STATSD_MAXUDPSIZE = 512

##################
# Custom headers #
##################
X_USER_ID_HEADER = 'X-Munch-User-Id'
X_POOL_HEADER = 'X-Munch-Pool'
X_MESSAGE_ID_HEADER = 'X-Munch-Message-Id'
X_HTTP_DSN_RETURN_PATH_HEADER = 'X-Munch-HTTP-Return-Path'
X_SMTP_DSN_RETURN_PATH_HEADER = 'X-Munch-SMTP-Return-Path'

#################
# Transactional #
#################
TRANSACTIONAL = {
    'X_USER_ID_HEADER': X_USER_ID_HEADER,
    'X_MESSAGE_ID_HEADER': X_MESSAGE_ID_HEADER,
    'X_HTTP_DSN_RETURN_PATH_HEADER': X_HTTP_DSN_RETURN_PATH_HEADER,
    'X_SMTP_DSN_RETURN_PATH_HEADER': X_SMTP_DSN_RETURN_PATH_HEADER,
    'X_MAIL_BATCH_HEADER': 'X-Munch-MailBatch',
    'X_MAIL_BATCH_CATEGORY_HEADER': 'X-Munch-MailBatch-Category',
    'X_MAIL_TRACK_OPEN_HEADER': 'X-Munch-Track-Open',
    'X_MAIL_TRACK_CLICKS_HEADER': 'X-Munch-Track-Clicks',
    'X_MAIL_UNSUBSCRIBE_HEADER': 'X-Munch-Unsubscribe',
    # Set these to drop to an unprivileged user after binding the SMTP daemon
    'DROP_PRIVILEGES_USER': None,
    'DROP_PRIVILEGES_GROUP': None,
    'EDGE_EHLO_AS': None,
    # Default is 200
    'EDGE_MAX_CONN': 200,
    'EDGE_TIMEOUTS': {'data_timeout': None, 'command_timeout': None},
    'PROXYPROTO_ENABLED': False,
    'STATUS_WORKER_QUEUE': 'munch.status',
    'SMTP_SMARTHOST_TLS': None,
    'SMTP_DSN_ADDRESS': "CHANGEME <foo@munch.example.com>",
    'HEADERS_TO_REMOVE': [X_MESSAGE_ID_HEADER],
    # Where do the smarthost binds
    'SMTP_BIND_HOST': '127.0.0.1',
    'SMTP_BIND_PORT': 1025,
    'SMTP_STOP_TIMEOUT': 5,
    'QUEUE_POLICIES': [
        'slimta.policy.split.RecipientSplit',
        'slimta.policy.split.RecipientDomainSplit',
        'munch.apps.transactional.policies.queue.bounces.Check',
        'munch.apps.transactional.policies.queue.sending_domain.Check',
        'slimta.policy.headers.AddDateHeader',
        'munch.apps.transactional.policies.queue.headers.AddMessageIdHeader',
        'slimta.policy.headers.AddReceivedHeader',
        'munch.apps.transactional.policies.queue.headers.Remove',
        'munch.apps.transactional.policies.queue.recipients.Clean',
        'munch.apps.transactional.policies.queue.identifier.Add',
        'munch.apps.transactional.policies.queue.store_mail.Store'],
    # Status logging
    'STATUS_WEBHOOK_RETRIES': 20,
    'STATUS_WEBHOOK_RETRY_INTERVAL': 180,
    'EXEC_QUEUE_POLICIES': [],
    # Filters applied to add custom headers to emails for each recipient
    # (order matters).
    'HEADERS_FILTERS': [
        'munch.apps.abuse.contentfilters.add_google_fbl_header'],
    'HEADERS_FILTERS_PARAMS': {
        # Values used to build the Google FBL header : X:X:X:tag
        # Search base for attributes is Mail object
        # 3 values max (order matters).
        'GOOGLE_FBL_TAGS': [
            'author.organization.pk',
            'batch.identifier',
        ],
    },
}

#############
# Campaigns #
#############
CAMPAIGNS = {
    'X_USER_ID_HEADER': X_USER_ID_HEADER,
    'X_MESSAGE_ID_HEADER': X_MESSAGE_ID_HEADER,
    # Cumulated max size of a message attachments
    # Applies on the actual files sizes but keep the b64-encoded size in mind.
    # Ex: 4000000 roughly equals 5333440 (~5MB) in Base64
    'ATTACHMENT_TOTAL_MAX_SIZE': 4000000,
    # Message attachment max size (per file) in bytes
    # Applies on the actual file size, not the
    # b64-encoded size which might be larger
    'ATTACHMENT_MAX_SIZE': 2097152,
    # How many mails can we add in a single API request ?
    'MAX_BULK_EMAILS': 10000,
    # Those may be set True for debug/testing only
    # On recipients adding they are checked for
    # an existing MX record on their domain.
    # BYPASS_DNS_CHECKS=True also disable this check
    'BYPASS_RECIPIENTS_MX_CHECK': False,
    'SKIP_SPAM_CHECK': False,
    'SKIP_VIRUS_CHECK': False,
    # When to opt-out after a bounce ?
    # Syntax is a 3-uplet :
    # - a list of matchs (matches beggining of bounce code)
    # - how many bounces before optout ?
    # - in which time frame are the bounces counted ?
    'BOUNCE_POLICY': [(['4.'], 10, 30 * 6), (['5.', ''], 3, 365)],
    # Filters applied to the template provided by the organization to produce
    # the HTML template (order matters).
    'HTML_TEMPLATE_FILTERS': [
        'munch.apps.campaigns.contentfilters.css_inline_html',
        'munch.apps.hosted.contentfilters.handle_images',
        'munch.apps.campaigns.contentfilters.clean_html'],
    # Filters applied on the HTML content build for each specific recipient
    # (order matters)
    'HTML_INDIVIDUAL_FILTERS': [
        'munch.apps.optouts.contentfilters.set_unsubscribe_url',
        'munch.apps.tracking.contentfilters.add_tracking_image',
        'munch.apps.tracking.contentfilters.rewrite_html_links',
        'munch.apps.hosted.contentfilters.set_web_link_url',
        'munch.apps.campaigns.contentfilters.apply_mailmerge'],
    # Filters applied on the plain text content build for each specific
    # recipient (order matters)
    'PLAINTEXT_INDIVIDUAL_FILTERS': [
        'munch.apps.tracking.contentfilters.rewrite_plaintext_links',
        'munch.apps.optouts.contentfilters.set_unsubscribe_url',
        'munch.apps.hosted.contentfilters.set_web_link_url',
        'munch.apps.campaigns.contentfilters.apply_mailmerge'],
    # Filters applied on the HTML content built for web (browser) view
    # (order matters).
    'WEB_HTML_INDIVIDUAL_FILTERS': [
        'munch.apps.tracking.contentfilters.web_rewrite_html_links',
        'munch.apps.campaigns.contentfilters.apply_mailmerge'],
    # Filters applied to add custom headers to emails for each recipient
    # (order matters).
    'HEADERS_FILTERS': [
        'munch.apps.abuse.contentfilters.add_google_fbl_header'],
    'HEADERS_FILTERS_PARAMS': {
        # Values used to build the Google FBL header : X:X:X:tag
        # Search base for attributes is Mail object
        # 3 values max (order matters).
        'GOOGLE_FBL_TAGS': [
            'message.author.organization.pk',
            'message.identifier',
        ],
    },
}

###############
# Backmuncher #
###############
BACKMUNCHER = {
    'SMTP_BIND_HOST': 'localhost',
    'SMTP_BIND_PORT': 12525,
    'EDGE_EHLO_AS': 'localhost',
    'DROP_PRIVILEGES_USER': None,
    'DROP_PRIVILEGES_GROUP': None,
}

#########
# Users #
#########
USERS = {
    # Value of CNAME record for organizations app_domains
    'ORGANIZATION_APP_DOMAIN_CNAME': 'munch.example.com'
}

###########
# Domains #
###########
DOMAINS = {
    # Should be included in the spf line
    'SPF_INCLUDE': 'munch.example.com',
    'DKIM_KEY_ID': 'munch',
    'DKIM_KEY_CONTENT': '"v=DKIM1; k=rsa; t=y; p=CHANGEME',
    'DKIM_CNAME': 'dkim.example.com'
}

############
# Optouts #
############
UNSUBSCRIBE_PLACEHOLDER = 'UNSUBSCRIBE_URL'
OPTOUTS = {
    'UNSUBSCRIBE_PLACEHOLDER': UNSUBSCRIBE_PLACEHOLDER,
}

##########
# Hosted #
##########
WEB_LINK_PLACEHOLDER = 'WEB_VERSION_URL'
HOSTED = {
    'WEB_LINK_PLACEHOLDER': WEB_LINK_PLACEHOLDER,
}

################
# Upload store #
################
UPLOAD_STORE = {
    'URL_PREFIX': 'uploads',
    'URL': 'http://munch.example.com',
    'IMAGE_MAX_WIDTH': 600,
    'BACKEND': 'munch.apps.upload_store.backends.LocalFileSystemStorage'
}

############
# Contacts #
############
CONTACTS = {
    # How many contacts can we add in a single API request ?
    'MAX_BULK_CONTACTS': 10000,
    'EXPIRATIONS': {
        'contact_queues:double-opt-in': timedelta(days=7),
        'contact_queues:bounce-check': timedelta(hours=1),
        'contact_queues:consumed_lifetime': timedelta(days=7),
        'contact_queues:failed_lifetime': timedelta(days=7),
        'contact_lists:double-opt-in': timedelta(days=7),
        'contact_lists:bounce-check': timedelta(hours=1),
        'contact_lists:consumed_lifetime': timedelta(days=7),
        'contact_lists:failed_lifetime': timedelta(days=7)}
}

############
# Mailsend #
############
MAILSEND = {
    # All Internal mailsend emails will be send to a blackhole
    'SANDBOX': True,
    'RELAY_TIMEOUTS': {
        'connect_timeout': 30.0, 'command_timeout': 30.0,
        'data_timeout': None, 'idle_timeout': None},
    # Timeout for MailStatus cache
    'MAILSTATUS_CACHE_TIMEOUT': 60 * 60 * 24 * 15,
    'X_POOL_HEADER': X_POOL_HEADER,
    'X_MESSAGE_ID_HEADER': X_MESSAGE_ID_HEADER,
    # Letting to None will make it use the host FQDN
    'SMTP_WORKER_EHLO_AS': 'munch.example.com',
    # Letting to None fallback to system routing
    # example: '1.2.3.4'
    'SMTP_WORKER_SRC_ADDR': None,
    # Backoff time is exponentially growing up to max_retry_interval and then
    # staying there on each retry till we reach time_before_drop.
    'RETRY_POLICY': {
        # Minimun time between two retries
        'min_retry_interval': 600,
        # Maximum time between two retries
        'max_retry_interval': 3600,
        # Time before we drop the mail and notify sender
        'time_before_drop': 2 * 24 * 3600},
    # Set this to an encoder fromhttps://docs.python.org/3.4/library/email.encoders.html#module-email.encoders  # noqa
    # to convert utf-8 emails to ascii
    'BINARY_ENCODER': None,
    'BLACKLISTED_HEADERS': [
        X_POOL_HEADER,
        X_HTTP_DSN_RETURN_PATH_HEADER,
        X_SMTP_DSN_RETURN_PATH_HEADER],
    'RELAY_POLICIES': [
        'munch.apps.transactional.policies.relay.headers.RewriteReturnPath',
        'munch_mailsend.policies.relay.headers.StripBlacklisted',
        # 'munch_mailsend.policies.relay.dkim.Sign'],
    ],
    'WORKER_POLICIES': [
        # 'munch_mailsend.policies.mx.pool.Policy',
        # 'munch_mailsend.policies.mx.rate_limit.Policy',
        # 'munch_mailsend.policies.mx.greylist.Policy',
        # 'munch_mailsend.policies.mx.warm_up.Policy',
    ],
    'WORKER_POLICIES_SETTINGS': {
        'rate_limit': {
            'domains': [
                (r'.*', 2)],
            'max_queued': 60 * 15},
        'warm_up': {
            'prioritize': 'equal',
            'domain_warm_up': {
                'matrix': [50, 100, 300, 500, 1000],
                'goal': 500,
                'max_tolerance': 10,
                'step_tolerance': 10,
                'days_watched': 10,
            },
            'ip_warm_up': {
                'matrix': [50, 100, 300, 500, 1000],
                'goal': 500,
                'max_tolerance': 10,
                'step_tolerance': 10,
                'enabled': False,
                'days_watched': 10,
            }
        },
        'pool': {'pools': ['default']},
        'greylist': {'min_retry': 60 * 5},
    },
    'WARM_UP_DOMAINS': {},
    'DKIM_PRIVATE_KEY': None,
    'DKIM_SELECTOR': None,
    'TLS': {'keyfile': None, 'certfile': None},
    'TASKS_SETTINGS': {
        'send_email': {
            'default_retry_delay': 180,
            'max_retries': (2 * 7 * 24 * 60 * 60) / 180
        },
        'route_envelope': {
            'default_retry_delay': 180,
            'max_retries': (2 * 7 * 24 * 60 * 60) / 180
        }
    }
}

#######
# API #
#######
SECURE_PROXY_SSL_HEADER = ('HTTP_X_FORWARDED_PROTOCOL', 'https')
JWT_AUTH = {
    'JWT_ALLOW_REFRESH': True,
    'JWT_PAYLOAD_HANDLER': 'munch.apps.users.jwt.jwt_payload_handler'
}
REST_FRAMEWORK = {
    'DEFAULT_VERSION': 'v1',
    'ALLOWED_VERSIONS': ['v1'],
    'DEFAULT_VERSIONING_CLASS': (
        'munch.core.utils.versioning.NamespaceVersioning'),
    'EXCEPTION_HANDLER': 'munch.core.utils.exceptions.exception_handler',
    'DEFAULT_PERMISSION_CLASSES': [
        'munch.core.utils.permissions.IsOwnerOrAdmin'],
    'DEFAULT_AUTHENTICATION_CLASSES': [
        'munch.apps.users.authentication.SecretBasicAuthentication',
        'rest_framework.authentication.SessionAuthentication',
        'rest_framework_jwt.authentication.JSONWebTokenAuthentication',
    ],
    'DEFAULT_RENDERER_CLASSES': (
        'rest_framework.renderers.JSONRenderer',
        'rest_framework.renderers.BrowsableAPIRenderer',
        'munch.core.utils.renderers.PaginatedCSVRenderer'
    ),
    'DEFAULT_PARSER_CLASSES': (
        'rest_framework.parsers.JSONParser',
        'rest_framework.parsers.FormParser',
        'rest_framework.parsers.MultiPartParser',
        'munch.core.utils.parsers.CSVParser',
    ),
    'DEFAULT_FILTER_BACKENDS': ['rest_framework.filters.DjangoFilterBackend'],
    'DEFAULT_PAGINATION_CLASS': (
        'munch.core.utils.pagination.CountedPageNumberPagination'),
    'VIEW_DESCRIPTION_FUNCTION': 'munch.core.utils.api.get_view_description',
    'TEST_REQUEST_RENDERER_CLASSES': (
        'rest_framework.renderers.MultiPartRenderer',
        'rest_framework.renderers.JSONRenderer',
        'rest_framework_csv.renderers.CSVRenderer',
    ),
}

#########
# Spamd #
#########
# host and port are resolved via DNS SRV
# Alternatively, you can define port and host (see commented lines)
# SPAMD_SERVICE_NAME = 'sa.example.munch'
SPAMD_PORT = 1783
SPAMD_HOST = 'localhost'

#########
# Clamd #
#########
# CLAMD_SERVICE_NAME = 'av.example.munch'
CLAMD_PORT = 49227
CLAMD_HOST = 'localhost'



import os
import json
import operator
from functools import wraps

from pygments import highlight
from pygments.lexers import JsonLexer
from pygments.formatters import HtmlFormatter
from django.conf import settings
from django.utils.safestring import mark_safe
from django.template.loader import get_template
from django.template import TemplateDoesNotExist
from django.urls import reverse
from django.urls.exceptions import NoReverseMatch


class NoDefaultProvided(object):
    pass


def rgetattr(obj, name, default=NoDefaultProvided):
    """
    Same as getattr(), but allows dot notation lookup
    See: http://stackoverflow.com/questions/11975781
    """

    try:
        return operator.attrgetter(name)(obj)
    except AttributeError:
        if default != NoDefaultProvided:
            return default
        raise


def template_exists(name):
    try:
        get_template(name)
        return True
    except TemplateDoesNotExist:
        return False


# We want to be able to specify two types of MUNCH_LOGIN_URL in settings:
#   - a django url, which can handle a reverse()
#   - an arbitrary string for external urls
# And since we generate password confirmations ourselves, we need this
# to properly set the context
def get_login_url():
    try:
        return reverse(settings.LOGIN_URL)
    except NoReverseMatch:
        return settings.LOGIN_URL


def get_mail_by_identifier(identifier, must_raise=True):
    from django.core.exceptions import ObjectDoesNotExist
    from munch.apps.campaigns.models import Mail as CampaignsMail
    from munch.apps.transactional.models import Mail as TransactionalMail

    if identifier.startswith('c-'):
        mail = CampaignsMail.objects.filter(identifier=identifier).first()
        if mail:
            return mail

    if identifier.startswith('t-'):
        mail = TransactionalMail.objects.filter(identifier=identifier).first()
        if mail:
            return mail

    if must_raise:
        raise ObjectDoesNotExist(
            'No Mail (campaigns.Mail, transactional.Mail) '
            'with identifier "{}".'.format(identifier))


def pretty_json_as_html(data, sort_keys=True):
    response = json.dumps(data, sort_keys=sort_keys, indent=2)
    # Truncate the data. Alter as needed
    response = response[:5000]

    # Get the Pygments formatter
    formatter = HtmlFormatter(style='colorful')

    # Highlight the data
    response = highlight(response, JsonLexer(), formatter)

    # Get the stylesheet
    style = "<style>" + formatter.get_style_defs() + "</style><br>"
    # Safe the output
    return mark_safe(style + response)


def get_worker_type():
    return os.environ.get('WORKER_TYPE', 'all').lower()


def monkey_patch_slimta_exception():
    from django.conf import settings

    if hasattr(settings, 'RAVEN_CONFIG'):
        import logging

        import slimta.logging  # noqa

        def log_sentry_exception(name, **kwargs):
            logger = logging.getLogger(name)
            logger.error('Unhandled exception', exc_info=True)

        def logline(log, type, typeid, operation, **data):
            if not data:
                log('{0}:{1}:{2}'.format(
                    type, typeid, operation), exc_info=True)
            else:
                data_str = ' '.join(
                    ['='.join((key, slimta.logging.log_repr.repr(val)))
                        for key, val in sorted(data.items())])
                log('{0}:{1}:{2} {3}'.format(
                    type, typeid, operation, data_str), exc_info=True)

        slimta.logging.log_exception = log_sentry_exception
        slimta.logging.logline = logline
        slimta.logging.socket_error_log_level = logging.WARNING


def save_timer(name):
    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            if settings.STATSD_ENABLED:
                from statsd.defaults.django import statsd
                timer = statsd.timer(name)
                timer.start()
            result = f(*args, **kwargs)
            if settings.STATSD_ENABLED:
                timer.stop()
            return result
        return wrapper
    return decorator



import email
import logging

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from slimta.queue import QueueError
from slimta.policy import QueuePolicy

# DSN tasks
from munch.apps.campaigns.tasks import handle_dsn as campaigns_dsn
from munch.apps.transactional.status import (
    handle_dsn_status as transactional_dsn)
# Unsubscribe tasks
from munch.apps.campaigns.tasks import handle_mail_optout
# Feedback loop tasks
from munch.apps.campaigns.tasks import handle_fbl

from munch.core.utils import get_mail_by_identifier
from munch.core.mail.models import return_path_parser
from munch.core.mail.models import unsubscribe_parser
from munch.core.mail.utils.parsers import extract_domain

log = logging.getLogger(__name__)


class MailHandler:
    def __init__(self, envelope, message):
        self.envelope = envelope
        self.message = message


class AbstractReportHandler(MailHandler):
    CONTENT_TYPE = 'multipart/report'

    @classmethod
    def can_handle(cls, message):
        return (
            (message.get_content_type() == cls.CONTENT_TYPE) and
            (message.get_param('report-type') == cls.REPORT_TYPE))

    def _get_part_by_types(self, types):
        """ Get the firt part of mail matcthing one of mimetypes

        @param types : a list/tuple of mimetypes
        @return headers dict
        """
        parts = self.message.walk()
        part = next((i for i in parts if i.get_content_type() in types), [])
        d = {}
        if part:
            for i in part.get_payload():
                if isinstance(i, email.message.Message):
                    d.update({k: str(v) for k, v in i.items()})
        return d

    def get_report(self):
        parttype = 'message/{}'.format(self.REPORT_TYPE)
        return self._get_part_by_types([parttype])

    def get_original(self):
        """ Gets the original message that caused the report.

        Note that only the headers part is supposed to be included.
        """
        parttypes = ('message/rfc822', 'text/rfc822-headers')
        return self._get_part_by_types(parttypes)


class DSNHandler(AbstractReportHandler):
    REPORT_TYPE = 'delivery-status'

    def apply(self):
        recipient = self.envelope.recipients[0]
        if not return_path_parser.is_valid(recipient):
            log.info('Invalid recipient: {}'.format(recipient))
            return None

        try:
            identifier = return_path_parser.get_uuid(recipient)
            get_mail_by_identifier(identifier)
        except (ValueError, ObjectDoesNotExist):
            log.info('Cannot find any mail with this recipient: {}'.format(
                recipient))
            return

        if identifier.startswith('c-'):
            return campaigns_dsn.apply_async(args=[
                self.message.as_string(), self.get_report()])
        elif identifier.startswith('t-'):
            return transactional_dsn.apply_async(args=[
                self.message.as_string(), self.get_report()])


class ARFHandler(AbstractReportHandler):
    REPORT_TYPE = 'feedback-report'

    def apply(self):
        original = self.get_original()
        return_path = original.get('Return-Path')
        if not return_path_parser.is_valid(return_path):
            log.info('Invalid Return-Path: {}'.format(return_path))
            return None

        try:
            identifier = return_path_parser.get_uuid(return_path)
            get_mail_by_identifier(identifier)
        except (ValueError, ObjectDoesNotExist):
            log.info('Cannot find any mail with this Return-Path: {}'.format(
                return_path))
            return

        original = self.get_original()
        if not original:
            log.info('Empty original message')
            return

        return handle_fbl.apply_async(args=[
            self.message.as_string(), original, self.get_report()])


class UnsubscribeHandler(MailHandler):
    def apply(self):
        recipient = self.envelope.recipients[0]
        if not unsubscribe_parser.is_valid(recipient):
            log.info('Invalid recipient: {}'.format(recipient))
            return None

        try:
            identifier = unsubscribe_parser.get_uuid(recipient)
            get_mail_by_identifier(identifier)
        except (ValueError, ObjectDoesNotExist):
            log.info('Cannot find any mail with this recipient: {}'.format(
                recipient))
            return

        return handle_mail_optout.apply_async(
            args=[dict(self.message)])

    @classmethod
    def can_handle(cls, msg):
        """
        Unsubscribe are just plain email, so there is no particular format.
        """
        return True


class BackMuncherQueuePolicy(QueuePolicy):
    def apply(self, envelope):
        if not envelope.recipients:
            raise QueueError('Missing recipient')
        recipient = envelope.recipients[0]

        if extract_domain(recipient) != settings.RETURNPATH_DOMAIN:
            raise QueueError('Domain not valid')

        message = email.message_from_bytes(
            envelope.flatten()[0] + envelope.flatten()[1])

        # Check recipient prefix
        prefix_allowed = False
        for prefix in ['return-', 'unsubscribe-', 'abuse']:
            if recipient.startswith(prefix):
                prefix_allowed = True
                break
        if not prefix_allowed:
            raise QueueError('Prefix not allowed')

        # Find handler
        handler_found = False
        for handler in [DSNHandler, ARFHandler, UnsubscribeHandler]:
            if handler.can_handle(message):
                log.info('Handler {} can handle'.format(handler.__name__))
                handler = handler(envelope, message)
                handler_found = True
                break
        if not handler_found:
            raise QueueError('No handler found')
        handler.apply()


class Queue:
    def __init__(self, relay):
        self.policy = BackMuncherQueuePolicy()

    def kill(self):
        pass

    def enqueue(self, envelope):
        try:
            envelope = self.policy.apply(envelope)
        except QueueError as exc:
            return [(envelope, exc)]
        return [(envelope, 0)]



import os
import email
from unittest import TestCase

from django.conf import settings
from slimta.queue import QueueError
from slimta.envelope import Envelope

from munch.apps.users.tests.factories import UserFactory
from munch.apps.campaigns.tests.factories import MailFactory
from munch.apps.campaigns.tests.factories import MessageFactory
from munch.apps.campaigns.models import get_base_mail_identifier
from munch.apps.transactional.models import get_mail_identifier

from ..backmuncher import DSNHandler
from ..backmuncher import ARFHandler
from ..backmuncher import UnsubscribeHandler
from ..backmuncher import BackMuncherQueuePolicy

UNSUBSCRIBE_EMAIL=open(os.path.join(
    os.path.dirname(__file__), 'samples/unsubscribe.eml')).read()
DSN_REPORT=open(os.path.join(
    os.path.dirname(__file__), 'samples/dsn_ok.eml')).read()
ARF_REPORT=open(os.path.join(
    os.path.dirname(__file__), 'samples/arf_abuse.eml')).read()


class TestSamples(TestCase):
    def test_multipart(self):
        arf = ARF_REPORT.replace('||FROM||', 'foo@bar')
        arf = arf.replace('||RETURNPATH||', 'foo@bar')
        dsn = DSN_REPORT.replace('||TO||', 'foo@bar')
        dsn = dsn.replace('||RETURNPATH||', 'foo@bar')
        self.assertTrue(email.message_from_string(arf).is_multipart())
        self.assertTrue(email.message_from_string(dsn).is_multipart())

    def test_multipart(self):
        dsn = DSN_REPORT.replace('||TO||', 'foo@bar')
        dsn = dsn.replace('||RETURNPATH||', 'foo@bar')
        m = email.message_from_string(dsn)
        self.assertIsNotNone(m['Subject'])
        self.assertTrue(m.is_multipart())


class RecipientTestCase(TestCase):
    def test_valid_recipient(self):
        recipient = 'foo@test.munch.example.com'
        dsn = DSN_REPORT.replace('||TO||', recipient)
        dsn = dsn.replace('||RETURNPATH||', 'foo@bar')
        message = email.message_from_string(dsn)
        envelope = Envelope()
        envelope.parse_msg(message)
        envelope.recipients = [recipient]
        with self.assertRaises(QueueError) as exc:
            BackMuncherQueuePolicy().apply(envelope)
        self.assertNotEqual(str(exc.exception), 'Missing recipient')

    def test_invalid_recipient(self):
        recipient = 'foo@test.munch.example.com'
        dsn = DSN_REPORT.replace('||TO||', recipient)
        dsn = dsn.replace('||RETURNPATH||', recipient)
        message = email.message_from_string(dsn)
        envelope = Envelope()
        envelope.parse_msg(message)
        with self.assertRaises(QueueError) as exc:
            BackMuncherQueuePolicy().apply(envelope)
        self.assertEqual(str(exc.exception), 'Missing recipient')


class ReturnPathDomainTestCase(TestCase):
    def test_valid_return_path(self):
        recipient = 'foo@test.munch.example.com'
        dsn = DSN_REPORT.replace('||TO||', recipient)
        dsn = dsn.replace('||RETURNPATH||', 'foo@bar')
        message = email.message_from_string(dsn)
        envelope = Envelope()
        envelope.parse_msg(message)
        envelope.recipients = [recipient]
        with self.assertRaises(QueueError) as exc:
            BackMuncherQueuePolicy().apply(envelope)
        self.assertNotEqual(str(exc.exception), 'Domain not valid')

    def test_invalid_return_path(self):
        recipient = 'foo@bar'
        dsn = DSN_REPORT.replace('||TO||', recipient)
        dsn = dsn.replace('||RETURNPATH||', recipient)
        message = email.message_from_string(dsn)
        envelope = Envelope()
        envelope.parse_msg(message)
        envelope.recipients = [recipient]
        with self.assertRaises(QueueError) as exc:
            BackMuncherQueuePolicy().apply(envelope)
        self.assertEqual(str(exc.exception), 'Domain not valid')


class PrefixTestCase(TestCase):
    def test_valid_prefix(self):
        recipient = 'return-foo@test.munch.example.com'
        dsn = DSN_REPORT.replace('||TO||', recipient)
        dsn = dsn.replace('||RETURNPATH||', 'foo@bar')
        message = email.message_from_string(dsn)
        envelope = Envelope()
        envelope.parse_msg(message)
        envelope.recipients = [recipient]
        BackMuncherQueuePolicy().apply(envelope)

    def test_invalid_prefix(self):
        recipient = 'foo@test.munch.example.com'
        dsn = DSN_REPORT.replace('||TO||', recipient)
        dsn = dsn.replace('||RETURNPATH||', recipient)
        message = email.message_from_string(dsn)
        envelope = Envelope()
        envelope.parse_msg(message)
        envelope.recipients = [recipient]
        with self.assertRaises(QueueError) as exc:
            BackMuncherQueuePolicy().apply(envelope)
        self.assertEqual(str(exc.exception), 'Prefix not allowed')


class DSNHandlerTestCase(TestCase):
    def test_parse(self):
        recipient = 'return-{}@test.munch.example.com'.format(
            get_mail_identifier())
        dsn = DSN_REPORT.replace('||TO||', recipient)
        dsn = dsn.replace('||RETURNPATH||', recipient)
        message = email.message_from_string(dsn)
        envelope = Envelope()
        envelope.parse_msg(message)
        envelope.recipients = [recipient]
        handler = DSNHandler(envelope, message)

        main_headers = email.message_from_string(handler.message.as_string())
        report_headers = handler.get_report()

        self.assertEqual(main_headers['To'], recipient)

        self.assertEqual(
            report_headers['Arrival-Date'],
            'Mon,  1 Sep 2014 17:36:45 +0000 (UTC)')
        self.assertEqual(report_headers['Diagnostic-Code'], 'smtp; 250 OK')
        self.assertEqual(report_headers['Status'], '2.0.0')
        self.assertEqual(report_headers['Action'], 'relayed')

    def test_invalide_recipient(self):
        recipient = 'return-foo@test.munch.example.com'
        dsn = DSN_REPORT.replace('||TO||', recipient)
        dsn = dsn.replace('||RETURNPATH||', 'foo@bar')
        message = email.message_from_string(dsn)
        envelope = Envelope()
        envelope.parse_msg(message)
        envelope.recipients = [recipient]
        handler = DSNHandler(envelope, message)
        self.assertIsNone(handler.apply())

    def test_unknown_identifier(self):
        recipient = 'return-{}@test.munch.example.com'.format(
            get_base_mail_identifier())
        dsn = DSN_REPORT.replace('||TO||', recipient)
        dsn = dsn.replace('||RETURNPATH||', 'foo@bar')
        message = email.message_from_string(dsn)
        envelope = Envelope()
        envelope.parse_msg(message)
        envelope.recipients = [recipient]
        handler = DSNHandler(envelope, message)
        self.assertIsNone(handler.apply())

        recipient = 'return-{}@test.munch.example.com'.format(
            get_mail_identifier())
        dsn = DSN_REPORT.replace('||TO||', recipient)
        dsn = dsn.replace('||RETURNPATH||', 'foo@bar')
        message = email.message_from_string(dsn)
        envelope = Envelope()
        envelope.parse_msg(message)
        envelope.recipients = [recipient]
        handler = DSNHandler(envelope, message)
        self.assertIsNone(handler.apply())

    def test_valid_dsn(self):
        user = UserFactory()
        message = MessageFactory(author=user)
        mail = MailFactory(message=message)

        recipient = 'return-{}@test.munch.example.com'.format(
            mail.identifier)
        dsn = DSN_REPORT.replace('||TO||', recipient)
        dsn = dsn.replace('||RETURNPATH||', recipient)
        message = email.message_from_string(dsn)
        envelope = Envelope()
        envelope.parse_msg(message)
        envelope.recipients = [recipient]
        handler = DSNHandler(envelope, message)
        self.assertIsNotNone(handler.apply())


class ARFHandlerTestCase(TestCase):
    def test_parse(self):
        recipient = 'return-{}@test.munch.example.com'.format(
            get_mail_identifier())
        arf = ARF_REPORT.replace('||TO||', recipient)
        arf = arf.replace('||RETURNPATH||', recipient)
        message = email.message_from_string(arf)
        envelope = Envelope()
        envelope.parse_msg(message)
        envelope.recipients = [recipient]
        handler = ARFHandler(envelope, message)

        main_headers = email.message_from_string(handler.message.as_string())
        orig_headers = handler.get_original()
        report_headers = handler.get_report()

        self.assertEqual(orig_headers['Return-Path'], recipient)
        self.assertEqual(report_headers['Feedback-Type'], 'abuse')
        self.assertEqual(main_headers['Date'], 'Thu, 8 Mar 2005 17:40:36 EDT')

    def test_invalide_returnpath(self):
        recipient = 'return-foo@test.munch.example.com'
        arf = ARF_REPORT.replace('||TO||', 'abuse@test.munch.example.com')
        arf = arf.replace('||RETURNPATH||', recipient)
        message = email.message_from_string(arf)
        envelope = Envelope()
        envelope.parse_msg(message)
        envelope.recipients = [recipient]
        handler = ARFHandler(envelope, message)
        self.assertIsNone(handler.apply())

    def test_unknown_identifier(self):
        recipient = 'return-{}@test.munch.example.com'.format(
            get_base_mail_identifier())
        arf = ARF_REPORT.replace('||TO||', 'abuse@test.munch.example.com')
        arf = arf.replace('||RETURNPATH||', recipient)
        message = email.message_from_string(arf)
        envelope = Envelope()
        envelope.parse_msg(message)
        envelope.recipients = [recipient]
        handler = ARFHandler(envelope, message)
        self.assertIsNone(handler.apply())

        recipient = 'return-{}@test.munch.example.com'.format(
            get_mail_identifier())
        arf = ARF_REPORT.replace('||TO||', 'abuse@test.munch.example.com')
        arf = arf.replace('||RETURNPATH||', recipient)
        message = email.message_from_string(arf)
        envelope = Envelope()
        envelope.parse_msg(message)
        envelope.recipients = [recipient]
        handler = ARFHandler(envelope, message)
        self.assertIsNone(handler.apply())

    def test_valid_arf(self):
        user = UserFactory()
        message = MessageFactory(author=user)
        mail = MailFactory(message=message)

        recipient = 'return-{}@test.munch.example.com'.format(
            mail.identifier)
        arf = ARF_REPORT.replace('||TO||', 'abuse@test.munch.example.com')
        arf = arf.replace('||RETURNPATH||', recipient)
        message = email.message_from_string(arf)
        envelope = Envelope()
        envelope.parse_msg(message)
        envelope.recipients = [recipient]
        handler = ARFHandler(envelope, message)
        self.assertIsNotNone(handler.apply())


class UnsubscribeHandlerTestCase(TestCase):
    def test_parse(self):
        recipient = 'unsubscribe-{}@test.munch.example.com'.format(
            get_mail_identifier())
        unsubscribe = UNSUBSCRIBE_EMAIL.replace('||TO||', recipient)
        unsubscribe = unsubscribe.replace('||RETURNPATH||', recipient)
        message = email.message_from_string(unsubscribe)
        envelope = Envelope()
        envelope.parse_msg(message)
        envelope.recipients = [recipient]
        handler = UnsubscribeHandler(envelope, message)

        main_headers = dict(handler.message)

        self.assertEqual(main_headers['To'], recipient)

    def test_invalide_returnpath(self):
        recipient = 'return-foo@test.munch.example.com'
        unsubscribe = UNSUBSCRIBE_EMAIL.replace('||TO||', recipient)
        unsubscribe = unsubscribe.replace('||RETURNPATH||', recipient)
        message = email.message_from_string(unsubscribe)
        envelope = Envelope()
        envelope.parse_msg(message)
        envelope.recipients = [recipient]
        handler = UnsubscribeHandler(envelope, message)
        self.assertIsNone(handler.apply())

    def test_unknown_identifier(self):
        recipient = 'return-{}@test.munch.example.com'.format(
            get_base_mail_identifier())
        unsubscribe = UNSUBSCRIBE_EMAIL.replace('||TO||', recipient)
        unsubscribe = unsubscribe.replace('||RETURNPATH||', recipient)
        message = email.message_from_string(unsubscribe)
        envelope = Envelope()
        envelope.parse_msg(message)
        envelope.recipients = [recipient]
        handler = UnsubscribeHandler(envelope, message)
        self.assertIsNone(handler.apply())

        recipient = 'return-{}@test.munch.example.com'.format(
            get_mail_identifier())
        unsubscribe = UNSUBSCRIBE_EMAIL.replace('||TO||', recipient)
        unsubscribe = unsubscribe.replace('||RETURNPATH||', recipient)
        message = email.message_from_string(unsubscribe)
        envelope = Envelope()
        envelope.parse_msg(message)
        envelope.recipients = [recipient]
        handler = UnsubscribeHandler(envelope, message)
        self.assertIsNone(handler.apply())

    def test_valid_unsubscribe(self):
        user = UserFactory()
        message = MessageFactory(author=user)
        mail = MailFactory(message=message)

        recipient = 'unsubscribe-{}@test.munch.example.com'.format(
            mail.identifier)
        unsubscribe = UNSUBSCRIBE_EMAIL.replace('||TO||', recipient)
        unsubscribe = unsubscribe.replace('||RETURNPATH||', recipient)
        message = email.message_from_string(unsubscribe)
        envelope = Envelope()
        envelope.parse_msg(message)
        envelope.recipients = [recipient]
        handler = UnsubscribeHandler(envelope, message)
        self.assertIsNotNone(handler.apply())



import logging

from django.core.management.base import BaseCommand

from munch.core.mail.utils import extract_domain

from ...models import Mail
from ...models import MailStatus

log = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Delete message ids. Theses emails won't be send."

    def add_arguments(self, parser):
        parser.add_argument('identifier', nargs='+', type=str)

    def handle(self, *args, **options):
        for identifier in options['identifier']:
            try:
                mail = Mail.objects.get(identifier=identifier)
                MailStatus.objects.create(
                    destination_domain=extract_domain(mail.recipient),
                    mail=mail, status=MailStatus.DELETED)
                self.stdout.write(self.style.SUCCESS(
                    "* {} (pk:{}) ignored".format(identifier, mail.pk)))
            except Mail.DoesNotExist:
                self.stdout.write(self.style.WARNING(
                    "* {} doesn't exist (ignored)".format(identifier)))
                exit(1)



import sys

from celery.signals import celeryd_after_setup

from munch.core.utils import get_worker_type
from munch.core.celery import catch_exception
from munch.core.celery import munch_tasks_router


def register_tasks():
    tasks_map = {
        'core': [
            'contacts.status.handle_bounce',
            'contacts.status.handle_bounce_expirations',
            'contacts.status.handle_failed_expirations',
            'contacts.status.handle_opt_ins_expirations',
            'contacts.status.handle_consumed_contacts_expirations'
        ]
    }
    munch_tasks_router.import_tasks_map(tasks_map, 'contacts')


@celeryd_after_setup.connect
@catch_exception
def configure_worker(instance, **kwargs):
    if get_worker_type() in ['core', 'all']:
        from .tasks import handle_bounce  # noqa
        from .tasks import handle_bounce_expirations  # noqa
        from .tasks import handle_failed_expirations  # noqa
        from .tasks import handle_opt_ins_expirations  # noqa
        from .tasks import handle_consumed_contacts_expirations  # noqa

        sys.stdout.write('[contacts-app] Registering worker as CORE...')
        munch_tasks_router.register_as_worker('core')



from abc import ABC

import django.core.mail.message
from django.conf import settings
from django.core.mail import EmailMultiAlternatives
from django.template.loader import render_to_string

from munch.core.utils import template_exists
from munch.apps.campaigns.contentfilters import css_inline_html

from . import mk_msgid


class ApplicationMessage(ABC):

    def __init__(self, to, subject, template, render_context={}):

        if not isinstance(to, list):
            to = [to]

        render_context.update({'application_message_recipient': to[0]})
        self.render_context = render_context
        body = render_to_string(template, render_context)

        # Use max line length from RFC2822 (78) instead of RFC5322 (998)
        # to force conversion to quoted-printable in almost all cases
        django.core.mail.message.RFC5322_EMAIL_LINE_LENGTH_LIMIT = 78

        self.msg = EmailMultiAlternatives(
            subject=subject, body=body, to=to, from_email=self.message_from,
            headers={
                'Auto-Submitted': 'auto-generated',
                'Return-Path': settings.SERVICE_MSG_RETURN_PATH,
                'Message-ID': mk_msgid()
            },
        )

    def add_html_part_if_exists(self, template, render_context=None):
        if render_context is None:
            render_context = self.render_context

        if template_exists(template):
            html = render_to_string(template, render_context)
            content = css_inline_html(html)
            self.msg.attach_alternative(content, 'text/html')

    def send(self):
        self.msg.send()


class ServiceMessage(ApplicationMessage):

    def __init__(self, *args, **kwargs):
        self.message_from = '{} <{}>'.format(
            settings.SERVICE_MSG_FROM_NAME, settings.SERVICE_MSG_FROM_EMAIL)
        super().__init__(*args, **kwargs)


class NotificationMessage(ApplicationMessage):

    def __init__(self, *args, **kwargs):
        self.message_from = '{} <{}>'.format(
            settings.NOTIFICATION_MSG_FROM_NAME,
            settings.NOTIFICATION_MSG_FROM_EMAIL)
        super().__init__(*args, **kwargs)
        if settings.NOTIFICATION_MSG_FROM_EMAIL != settings.SERVICE_MSG_FROM_EMAIL:  # noqa
            self.msg.extra_headers.update({
                'Reply-To': settings.SERVICE_MSG_FROM_EMAIL})



from django.urls import reverse
from django.contrib import admin
from django.db.models import Count

from munch.core.models import Category
from munch.apps.campaigns.models import Message
from munch.apps.transactional.models import MailBatch


class MessageInline(admin.TabularInline):
    extra = 0
    max_num = 0
    model = Message
    fields = [
        'sender_email', 'sender_name', 'status', 'creation_date',
        'track_clicks', 'track_open', 'is_spam']
    readonly_fields = [
        'sender_email', 'sender_name', 'status', 'creation_date',
        'track_clicks', 'track_open', 'is_spam']


class MailBatchInline(admin.TabularInline):
    extra = 0
    max_num = 0
    model = MailBatch
    fields = ['name', 'creation_date']
    readonly_fields = ['name', 'creation_date']


class CategoryAdmin(admin.ModelAdmin):
    list_display = [
        'name', 'show_message_count', 'show_batch_count',
        'author_url', 'organization_url']
    search_fields = [
        'name', 'author__identifier', 'author__organization__name']
    readonly_fields = ['author_url']
    inlines = [MessageInline, MailBatchInline]
    raw_id_fields = ['author']

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.annotate(
            batch_count=Count('mailbatches'),
            message_count=Count('messages'))

    def author_url(self, instance):
        return '<a target="_blank" href="{}">{}</a>'.format(
            reverse('admin:users_munchuser_change', args=[instance.author.pk]),
            instance.author)
    author_url.allow_tags = True
    author_url.short_description = 'Author'

    def organization_url(self, instance):
        return '<a target="_blank" href="{}">{}</a>'.format(
            reverse(
                'admin:users_organization_change',
                args=[instance.author.organization.pk]),
            instance.author.organization.name)
    organization_url.allow_tags = True
    organization_url.short_description = 'Organization'

    def show_message_count(self, instance):
        return instance.message_count
    show_message_count.admin_order_field = 'message_count'
    show_message_count.short_description = 'Messages Count'

    def show_batch_count(self, instance):
        return instance.batch_count
    show_batch_count.admin_order_field = 'batch_count'
    show_batch_count.short_description = 'Batches Count'


admin.site.register(Category, CategoryAdmin)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.3 on 2016-06-02 09:13
from __future__ import unicode_literals

import django.contrib.postgres.fields.hstore
from django.db import migrations, models
from django.contrib.postgres.operations import HStoreExtension


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        HStoreExtension(),
        migrations.CreateModel(
            name='LinkMap',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('identifier', models.CharField(db_index=True, max_length=50, unique=True, verbose_name='identifier')),
                ('link', models.URLField(verbose_name='link', max_length=500)),
            ],
        ),
        migrations.CreateModel(
            name='TrackRecord',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('identifier', models.CharField(db_index=True, max_length=150, verbose_name='identifier')),
                ('kind', models.CharField(db_index=True, max_length=50, verbose_name='kind')),
                ('properties', django.contrib.postgres.fields.hstore.HStoreField(blank=True, null=True, verbose_name='properties')),
                ('creation_date', models.DateTimeField(auto_now_add=True, verbose_name='creation date')),
            ],
            options={
                'ordering': ['creation_date'],
            },
        ),
    ]



from django.conf.urls import url
from django.conf.urls import include

from munch.core.utils.routers import APIRouter
from munch.urls import api_urlpatterns_v1

from . import views

router = APIRouter()
router.register('messages', views.MessageViewSet, base_name='message')
router.register('recipients', views.MailViewSet, base_name='recipient')
router.register('bounces', views.BouncesViewSet, base_name='bounce')
router.register(
    'attachments',
    views.MessageAttachmentViewset, base_name='messageattachment')

urlpatterns = [
    url('^messages/(?P<msg_pk>\d+)/preview_send/$',
        views.PreviewSendMessageView.as_view()),
    url('^recipients/(?P<mail_pk>\d+)/optout/$',
        views.MailOptOutView.as_view()),
] + router.urls

api_urlpatterns_v1 += [url('', include(urlpatterns, namespace='campaigns'))]



from .__about__ import (
    __author__, __copyright__, __email__, __license__, __summary__, __title__,
    __uri__, __version__
)

__all__ = [
    "__title__", "__summary__", "__uri__", "__version__", "__author__",
    "__email__", "__license__", "__copyright__",
]



import re
import magic
import base64
import hashlib
import mimetypes
import urllib
import urllib.request

from django.core.files.base import ContentFile

from munch.apps.campaigns.exceptions import InvalidSubmitedData
from munch.apps.upload_store.models import Image

from .exceptions import TooBigMedia
from .exceptions import InvalidMimeType

mimetypes.init()


class HostedMaterial:
    """Represents a stored file (abstract classs) """
    MAX_SIZE = 1024 * 1024 * 4  # 4MiO

    def __init__(self, organization):
        self.organization = organization

    def store(self):
        if len(self.data) >= self.MAX_SIZE:
            raise TooBigMedia(self.identifying_name, self.MAX_SIZE)

        mime = magic.from_buffer(self.data, mime=True)
        if mime not in self.allowed_mimetypes:
            raise InvalidMimeType(mime)

        self.extension = mimetypes.guess_extension(mime)

        # weirdness from mimetypes
        if self.extension == '.jpe':
            self.extension = '.jpeg'

        checksum = hashlib.sha1(self.data).hexdigest()
        fn = '{}{}'.format(checksum, self.extension)

        img = Image(organization=self.organization)
        img.file.save(fn, ContentFile(self.data))
        return img.get_absolute_url()


class AbstractHostedImage(HostedMaterial):
    allowed_mimetypes = (
        'image/jpeg', 'image/gif', 'image/png', 'image/svg+xml')


class InlineImage(AbstractHostedImage):
    """ Stores an image from a data-uri XML value """
    r_data_uri = re.compile(
        r'data:(?P<mimetype>image/(jpeg|png|gif));base64,(?P<data>.*)')

    def __init__(self, data_uri, *args, **kwargs):
        """
        @param data_uri : the data-uri string
        example of valid data_uri is :

        data:image/gif;base64,R0lGODdhAQABAIABAOvr6////ywAAAAAAQABAAACAkQBADs=
        """
        super().__init__(*args, **kwargs)

        self.identifying_name = data_uri
        if len(data_uri) > 50:
            self.identifying_name = data_uri[:50] + '...'
        else:
            self.identifying_name = data_uri

        m = self.r_data_uri.match(data_uri)
        if m:
            self.data = base64.b64decode(m.group('data').encode())
        else:
            raise InvalidSubmitedData('not a data uri : {}'.format(data_uri))


class HostedImage(AbstractHostedImage):
    """ Download and stores an image from an url """
    def __init__(self, http_uri, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.identifying_name = http_uri
        response = urllib.request.urlopen(http_uri)
        self.data = response.read(self.MAX_SIZE)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2016-09-26 16:10
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('users', '0004_remove_organization_password_reset_link'),
    ]

    operations = [
        migrations.AlterField(
            model_name='munchuser',
            name='first_name',
            field=models.CharField(max_length=50, verbose_name='first name'),
        ),
        migrations.AlterField(
            model_name='munchuser',
            name='last_name',
            field=models.CharField(max_length=50, verbose_name='last name'),
        ),
    ]



from rest_framework import serializers
from rest_framework.serializers import ValidationError

from munch.core.utils.serializers import HALLinksField
from munch.core.utils.serializers import CurrentOrganizationDefault

from ...models import SendingDomain


class AltOrganizationsField(serializers.HyperlinkedRelatedField):
    def get_queryset(self, *args, **kwargs):
        return self.context.get('request').user.organization.children.all()


class SendingDomainSerializer(serializers.HyperlinkedModelSerializer):
    dkim_status = serializers.ReadOnlyField(allow_null=True)
    app_domain_status = serializers.ReadOnlyField(allow_null=True)
    _links = HALLinksField(
        nested_endpoints=['revalidate'],
        view_name='domains:sendingdomain-detail')
    alt_organizations = AltOrganizationsField(
        many=True, view_name='users:organization-detail', required=False)
    organization = serializers.HiddenField(
        default=serializers.CreateOnlyDefault(CurrentOrganizationDefault()))

    class Meta:
        model = SendingDomain
        fields = [
            'url', 'name', 'dkim_status', 'app_domain_status',
            'organization', 'alt_organizations', 'app_domain',
            'app_domain_status_date', 'dkim_status_date', '_links']
        read_only_fields = [
            'organization', 'dkim_status_date', 'app_domain_status_date']
        extra_kwargs = {'url': {'view_name': 'domains:sendingdomain-detail'}}

    def validate(self, attrs):
        for organization in attrs.get('alt_organizations', []):
            if organization.parent_id != attrs.get('organization').id:
                raise ValidationError('Invalid organization')
        return attrs



from django.test import TestCase
from django.conf import settings
from rest_framework.test import APIClient

from munch.apps.users.tests.factories import UserFactory
from munch.apps.campaigns.tests.factories import MessageFactory

from ...models import Category
from ...tests.factories import CategoryFactory

class CategoryAPITestCase(TestCase):
    def setUp(self):
        self.api_prefix = 'v1'
        self.user = UserFactory(groups=['administrators'])

        self.client = APIClient()
        self.client.login(identifier=self.user.identifier, password='password')

        self.html = '<html>{}</html>'.format(
            settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER'])

    def test_can_delete_empty_category(self):
        category = CategoryFactory(author=self.user)
        self.assertEqual(Category.objects.count(), 1)
        resp = self.client.delete('/{}/categories/{}/'.format(
            self.api_prefix, category.pk))
        self.assertEqual(resp.status_code, 204)
        self.assertEqual(Category.objects.count(), 0)

    def test_cannot_delete_non_empty_category(self):
        category = CategoryFactory(author=self.user)

        MessageFactory(category=category, author=self.user)

        self.assertEqual(Category.objects.count(), 1)
        self.assertEqual(Category.objects.first().messages.count(), 1)
        resp = self.client.delete('/{}/categories/{}/'.format(
            self.api_prefix, category.pk))
        self.assertEqual(resp.status_code, 403)
        self.assertEqual(Category.objects.count(), 1)
        self.assertEqual(Category.objects.first().messages.count(), 1)

    def test_cannot_delete_not_owned_category(self):
        user = UserFactory(groups=['administrators'])

        category = CategoryFactory(author=user)

        self.assertEqual(Category.objects.count(), 1)
        resp = self.client.delete('/{}/categories/{}/'.format(
            self.api_prefix, category.pk))
        self.assertEqual(resp.status_code, 403)
        self.assertEqual(Category.objects.count(), 1)



from django.db import models
from django.db.models import Q
from django.utils.translation import ugettext_lazy as _

from .managers import OwnedModelQuerySet
from .managers import OptionallyOwnedModelQuerySet


class OwnedModelMixin:
    """ Defines the relationship to an owner and an author

    Inheriting model should define two attributes in django queryset format.

    * `owner_path` points to a users.models.Organization
    * `author_path` points to a users.models.MunchUser

    Examples:
      * `owner_path='foo'` states that the owner is in self.foo
      * `owner_path='foo__bar'` states that the owner is in self.foo.bar
      * `owner_path='self'` states that the owner is the object itself


    If you want to add the extra permissions
        (related to author and organization),
    you have to explicitly inherit metaclass.

    Example usage:

        class SomeClass(models.Model, OwnedModelMixin):
            class Meta(OwnedModelMixin.Meta):
                pass

    """
    # owner_path = 'foo__bar'

    # just a way to mark that there is no path to author
    IRRELEVANT = 'irrelevant'

    class Meta:
        default_permissions = (
            'view', 'add', 'change', 'delete',
            'view_mine', 'change_mine', 'delete_mine',
            'view_organizations', 'change_organizations',
            'delete_organizations')

    def __init__(self, *args, **kwargs):
        if not hasattr(self, 'owner_path'):
            raise NotImplementedError(
                ('{}, as a OwnedModelMixin implementor, should define a' +
                    ' "owner_path" attribute').format(self.__class__))

    @classmethod
    def defines_author(cls):
        return cls.author_path != cls.IRRELEVANT

    def get_owner(self):
        """
        :rtype Organization:
        """
        o = self

        if self.owner_path != 'self':
            for i in self.owner_path.split('__'):
                o = getattr(o, i)
        return o

    def get_author(self):
        """
        :rtype MunchUser:
        """
        o = self

        if self.author_path != 'self' and self.author_path != self.IRRELEVANT:
            for i in self.author_path.split('__'):
                o = getattr(o, i)
        return o

    @classmethod
    def has_owner_direct_link(cls):
        return not ('__' in cls.owner_path) and (cls.owner_path != 'self')

    @classmethod
    def has_author_direct_link(cls):
        return not ('__' in cls.author_path) and (cls.author_path != 'self')

    @classmethod
    def is_author_class(cls):
        return cls.author_path == 'self'

    @classmethod
    def is_owner_class(cls):
        return cls.owner_path == 'self'

    @staticmethod
    def resolve_serializer_path(serializer_data, path):
        """ Given a django queryset filter, get the object following relations,
        starting from a serializer, not an instance

        ex:
        resolve_path(o, 'foo__bar') will fetch o['foo'].bar
        resolve_path(o, 'self') will fetch o
        """
        if path == 'self':
            raise ValueError('Cannot resolve "self"')
        else:
            path_bits = path.split('__')
            # first level of relation is dict-like (nb: it's a serializer)
            o = serializer_data[path_bits[0]]
            # following levers is normal model attributes traversing
            for path_bit in path_bits[1:]:
                o = getattr(o, path_bit)
        return o

    @classmethod
    def get_directly_linked_obj(cls, serializer_data):
        """
        In case there is a relation chain to author, gets the closest instance
        to the author in the chain.

        ex: if called with data of a mail and we have the Mail-Message-User
        chain, returns the message instance.
        """
        if cls.has_author_direct_link() or cls.is_author_class():
            raise ValueError(
                'the author has direct link to this class or is this class')
        else:
            path_bits = cls.author_path.split('__')
            # first level of relation is dict-like (nb: it's a serializer)
            o = serializer_data[path_bits[0]]
            # following levers is normal model attributes traversing
            for path_bit in path_bits[1:-1]:
                o = getattr(o, path_bit)
        return o


class OptionnallyOwnedModelMixin(OwnedModelMixin):
    """ A special case of OwnedModelMixin, where ownership of an object is
    optionnal. The queryset should return owned objects and orphans.
    """

    @classmethod
    def owned_qs(cls, qs, owner):
        if cls.owner_path == 'self':
            return qs.filter(Q(pk=owner.pk) | Q(pk__isnull=True))
        else:
            return qs.filter(
                Q(**{cls.owner_path: owner}) | Q(
                    **{cls.owner_path + '__isnull': True}))

    def is_public(self):
        return self.get_author() is None or self.get_owner() is None

    # for admin interface
    is_public.boolean = True
    is_public.short_description = _('public')


class AbstractOwnedModel(models.Model, OwnedModelMixin):
    class Meta(OwnedModelMixin.Meta):
        abstract = True

    objects = OwnedModelQuerySet.as_manager()


class AbstractOptionnallyOwnedModel(models.Model, OptionnallyOwnedModelMixin):
    class Meta(OptionnallyOwnedModelMixin.Meta):
        abstract = True

    objects = OptionallyOwnedModelQuerySet.as_manager()



import logging

from unittest.mock import patch
from datetime import timedelta

from django.conf import settings
from django.test import TestCase
from django.test.utils import override_settings
from libfaketime import fake_time

from munch.apps.campaigns.models import Mail
from munch.apps.campaigns.models import MailStatus
from munch.apps.campaigns.tests.factories import MailFactory
from munch.apps.campaigns.tests.factories import MessageFactory
from munch.apps.campaigns.tests.factories import MailStatusFactory
from munch.apps.users.tests.factories import UserFactory

from munch.apps.spamcheck.tests import get_spam_result_mock

from .utils import WebKey
from .views import TRACKER_PIXEL
from .models import LinkMap
from .models import TrackRecord
from .models import READ_BROWSER
from .models import READ_MUA_PIXEL

from .contentfilters import LinksRewriter
from .contentfilters import WebVersionLinksRewriter
from .contentfilters import rewrite_plaintext_links


logging.disable(logging.CRITICAL)


class TestTrackingPixel(TestCase):
    def setUp(self):
        self.user = UserFactory()

    def test_valid_tracker(self):
        with fake_time('2016-10-10 08:00:00'):
            message = MessageFactory(status='message_ok', author=self.user)
            mail = MailFactory(message=message)
            MailStatusFactory(mail=mail)
            MailStatusFactory(mail=mail, status=MailStatus.SENDING)
            MailStatusFactory(mail=mail, status=MailStatus.DELIVERED)
        with fake_time('2016-10-10 08:00:15'):
            resp = self.client.get('/t/open/{}'.format(mail.identifier))

        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp['Content-Type'], 'image/gif')
        self.assertEqual(resp.content, TRACKER_PIXEL)

        records = TrackRecord.objects.filter(
            identifier=mail.identifier,
            kind='read', properties__source=READ_MUA_PIXEL)

        self.assertEqual(records.count(), 1)

    def test_unfilled_tracker(self):
        resp = self.client.get('/t/open/')
        self.assertEqual(resp.status_code, 404)

    def test_invalid_tracker(self):
        resp = self.client.get('/t/open/4242')
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp['Content-Type'], 'image/gif')
        self.assertEqual(resp.content, TRACKER_PIXEL)

    def test_invalid_method(self):
        message = MessageFactory(status='message_ok', author=self.user)
        resp = self.client.post('/t/open/{}'.format(message.identifier))
        self.assertEqual(resp.status_code, 405)

    def test_reaction_time(self):
        with fake_time('2016-10-10 08:00:00'):
            message = MessageFactory(status='message_ok', author=self.user)
            mail = MailFactory(message=message)
            MailStatusFactory(mail=mail)
            MailStatusFactory(mail=mail, status=MailStatus.SENDING)
            MailStatusFactory(mail=mail, status=MailStatus.DELIVERED)
        with fake_time('2016-10-10 08:00:15'):
            self.client.get('/t/open/{}'.format(mail.identifier))
            records = TrackRecord.objects.filter(
                identifier=mail.identifier, kind='read')

        self.assertEqual(records.first().properties['reaction_time'], '15')

    def test_html_generation(self):
        message = MessageFactory(
            author=self.user,
            track_open=True, html='<body><h1>Foo</h1>{}</body>'.format(
                settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER']))
        mail = MailFactory(message=message)

        with patch(
                'munch.apps.spamcheck.SpamChecker.check',
                side_effect=get_spam_result_mock):
            message.save()

        content = message.to_mail(mail)
        self.assertIn('t/open/', content.alternatives[0][0])

    def test_html_non_generation(self):
        message = MessageFactory(
            author=self.user,
            track_open=False, html='<body><h1>Foo</h1>{}</body>'.format(
                settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER']))
        with patch(
                'munch.apps.spamcheck.SpamChecker.check',
                side_effect=get_spam_result_mock):
            message.save()

        mail = MailFactory(message=message)
        content = message.to_mail(mail)
        self.assertNotIn('pixel.gif', content.alternatives[0][0])

    def test_html_generation_preserve_doctype(self):
        message = MessageFactory(
            author=self.user, track_open=True, html=(
                '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//E'
                'N" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">'
                '<html xmlns="http://www.w3.org/1999/xhtml">{}</html>').format(
                    settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER']))

        with patch(
                'munch.apps.spamcheck.SpamChecker.check',
                side_effect=get_spam_result_mock):
            message.save()

        mail = MailFactory(message=message)
        content = message.to_mail(mail)
        self.assertIn(
            '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" '
            '"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">',
            content.alternatives[0][0])


class TestWebTracking(TestCase):
    def setUp(self):
        self.user = UserFactory()
        self.message = MessageFactory(author=self.user)

    def test_hosted_mail_nokey_notrack(self):
        self.client.get('/archive/{}/'.format(self.message.identifier))
        self.assertEqual(TrackRecord.objects.filter(kind='read').count(), 0)

    def test_hosted_mail_badkey_notrack(self):
        self.client.get('/archive/{}/?web_key=1234'.format(
            self.message.identifier))
        self.assertEqual(TrackRecord.objects.filter(kind='read').count(), 0)

    def test_hosted_mail_with_key_works_tracks(self):
        mail = MailFactory(message=self.message)
        MailStatusFactory(mail=mail, status=MailStatus.QUEUED)
        MailStatusFactory(mail=mail, status=MailStatus.SENDING)
        MailStatusFactory(mail=mail, status=MailStatus.DELIVERED)

        wk = WebKey.from_instance(Mail.objects.first().identifier)
        response = self.client.get('/archive/{}/?web_key={}'.format(
            self.message.identifier, wk.token))
        self.assertEqual(response.status_code, 200)

        record = TrackRecord.objects.filter(kind='read').first()
        self.assertEqual(TrackRecord.objects.filter(kind='read').count(), 1)
        self.assertEqual(record.properties['source'], READ_BROWSER)


class TestTrackingLinks(TestCase):
    def setUp(self):
        self.user = UserFactory()
        self.message = MessageFactory(
            author=self.user, track_clicks=True, track_open=False,
            html=(
                '<body><a href="http://example.com">'
                'Hi</a>{}</body>').format(
                    settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER']))
        with patch(
                'munch.apps.spamcheck.SpamChecker.check',
                side_effect=get_spam_result_mock):
            self.message.save()
        self.mail = MailFactory(message=self.message)

    def test_html_generation(self):
        with patch(
                'munch.apps.spamcheck.SpamChecker.check',
                side_effect=get_spam_result_mock):
            self.message.save()
        content = self.message.to_mail(self.mail)
        self.assertIn('/clicks/m/', content.alternatives[0][0])

    def test_plaintext_generation(self):
        with patch(
                'munch.apps.spamcheck.SpamChecker.check',
                side_effect=get_spam_result_mock):
            self.message.save()

        content = self.message.to_mail(self.mail)
        self.assertIn('/clicks/m/', content.body)

    def test_html_no_unsubscribe_or_viewonline_rewrite(self):
        self.message.track_clicks = True
        self.message.html = '<body><h1>Hi</h1>{}</body>'.format(
            settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER'])
        with patch(
                'munch.apps.spamcheck.SpamChecker.check',
                side_effect=get_spam_result_mock):
            self.message.save()

        mail = MailFactory(message=self.message)
        content = self.message.to_mail(mail)
        self.assertNotIn('/clicks/m/', content.alternatives[0][0])

    def test_mk_redirection(self):
        mail = MailFactory(message=self.message)
        url = 'http://example.com'
        redir_url = WebVersionLinksRewriter.rewrite(
            mail.identifier,
            self.message.get_app_url(),
            mail.message.msg_links,
            url)

        # build local url from fully qualified url
        local_redir_url = redir_url.split('://')[1].split('/', 1)[1]

        resp = self.client.get('/' + local_redir_url)
        self.assertEqual(resp.status_code, 302)
        self.assertEqual(resp['Location'], url)

    def test_mk_web_redirection(self):
        mail = MailFactory(message=self.message)
        url = 'http://example.com'
        redir_url = WebVersionLinksRewriter.rewrite(
            mail.identifier,
            self.message.get_app_url(),
            mail.message.msg_links,
            url)

        # build local url from fully qualified url
        local_redir_url = redir_url.split('://')[1].split('/', 1)[1]

        self.assertEqual(LinkMap.objects.count(), 1)
        self.assertEqual(TrackRecord.objects.filter(
            kind='click', identifier=mail.identifier).count(), 0)
        resp = self.client.get('/' + local_redir_url)
        self.assertEqual(resp.status_code, 302)
        self.assertEqual(resp['Location'], url)
        self.assertEqual(TrackRecord.objects.filter(
            kind='click', identifier=mail.identifier).count(), 1)

    def test_invalid_tracker(self):
        resp = self.client.get(
            '/t/clicks/m/0202020202020202/CCCCCDDDDD')
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.content, b'No URL found with this ID')

    def test_html_generation_preserve_doctype(self):
        self.message.track_clicks = True
        self.message.html = (
            '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" '
            '"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">'
            '<html xmlns="http://www.w3.org/1999/xhtml">{}</html>').format(
                settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER'])
        with patch(
                'munch.apps.spamcheck.SpamChecker.check',
                side_effect=get_spam_result_mock):
            self.message.save()

        mail = MailFactory(message=self.message)
        content = self.message.to_mail(mail)
        self.assertIn(
            (
                '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//'
                'EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.'
                'dtd">'),
            content.alternatives[0][0])

    def test_web_version_links_can_track(self):
        self.message.track_clicks = True
        self.message.save()

        mail = MailFactory(message=self.message)

        web_token= WebKey.from_instance(mail.identifier)
        response = self.client.get(
            '/archive/{}/?web_key={}'.format(
                self.message.identifier, web_token.token))

        self.assertIn(b'/t/clicks/w/', response.content)

    def test_web_version_links_can_no_track(self):
        self.message.track_clicks = False
        self.message.save()

        mail = MailFactory(message=self.message)
        web_key = WebKey.from_instance(mail.identifier)
        response = self.client.get(
            '/archive/{}/?web_key={}'.format(
                self.message.identifier, web_key.token))

        self.assertNotIn(b'/t/clicks/w/', response.content)

    def test_web_versions_open_can_no_track(self):
        self.message.track_open = False
        self.message.save()
        mail = MailFactory(message=self.message)
        self.assertNotIn('web_key', mail.web_view_url)

    def test_web_versions_open_can_track(self):
        self.message.track_open = True
        self.message.save()
        mail = MailFactory(message=self.message)
        self.assertIn('web_key', mail.web_view_url)


@override_settings(SECRET_KEY='123412341234')
class TestWebKey(TestCase):
    def setUp(self):
        self.user = UserFactory()
        self.message = MessageFactory(author=self.user)
        self.mail = MailFactory(message=self.message)

    def test_mk_web_key(self):
        key = WebKey.from_instance(self.mail.identifier)
        self.assertIsInstance(key.token, str)

    def test_check_web_key(self):
        key = WebKey.from_instance(self.mail.identifier)

        key2 = WebKey(key.token)
        identifier = key2.get_identifier()
        self.assertEqual(identifier, self.mail.identifier)

    def test_check_web_key_bad_sig(self):
        # Sign with a bad key
        with override_settings(SECRET_KEY='2121'):
            web_key = WebKey.from_instance(self.mail.identifier)

        web_key2 = WebKey(web_key.token)
        identifier = web_key2.get_identifier()
        self.assertIsNone(identifier)


class TestContentFilters(TestCase):
    def setUp(self):
        self.user = UserFactory()
        self.message = MessageFactory(
            author=self.user, track_clicks=True, track_open=False,
            msg_links={'AAAAABBBBB': 'http://example.com'},
            html='<body><a href="http://example.com">Hi</a>{}</body>'.format(
                settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER']))
        with patch(
                'munch.apps.spamcheck.SpamChecker.check',
                side_effect=get_spam_result_mock):
            self.message.save()

    def test_rewrite_footlink(self):
        text = """Hi
        We are [1]

        [1]: http://example.com
        """
        mail = MailFactory(message=self.message)
        out = rewrite_plaintext_links(
            text,
            app_url=self.message.get_app_url(),
            unsubscribe_url=mail.unsubscribe_url,
            mail_identifier=mail.identifier,
            links_map=self.message.msg_links)
        self.assertIn('/clicks/m/', out)

    def test_dont_rewrite_bodylink(self):
        text = """Hi

        Go to http://example.com
        """
        mail = MailFactory(message=self.message)
        out = rewrite_plaintext_links(
            text,
            app_url=self.message.get_app_url(),
            unsubscribe_url=mail.unsubscribe_url,
            mail_identifier=mail.identifier,
            links_map=self.message.msg_links)
        self.assertNotIn('/clicks/m', out)
